from PyQt6 import QtCore

from vorta.borg.extract import BorgExtractJob
from vorta.borg.list_archive import BorgListArchiveJob
from vorta.store.models import ArchiveModel
from vorta.utils import choose_file_dialog
from vorta.views.dialogs.archive import extract as extract_dialog
from vorta.views.dialogs.archive.extract import ExtractDialog, ExtractTree


class ArchiveExtract:
    def __init__(self, tab):
        self.tab = tab

    def extract_action(self):
        """
        Open a dialog for choosing what to extract from the selected archive.
        """
        profile = self.tab.profile()

        row_selected = self.tab.archiveTable.selectionModel().selectedRows()
        if row_selected:
            archive_cell = self.tab.archiveTable.item(row_selected[0].row(), 4)
            if archive_cell:
                archive_name = archive_cell.text()
                params = BorgListArchiveJob.prepare(profile, archive_name)

                if not params['ok']:
                    self.tab._set_status(params['message'])
                    return
                self.tab._set_status('')
                self.tab._toggle_all_buttons(False)

                job = BorgListArchiveJob(params['cmd'], params, self.tab.profile().repo.id)
                job.updated.connect(self.tab.mountErrors.setText)
                job.result.connect(self.extract_list_result)
                self.tab.app.jobs_manager.add_job(job)
                return job
        else:
            self.tab._set_status(self.tab.tr('Select an archive to restore first.'))

    def extract_list_result(self, result):
        """Process the contents of the archive to extract."""
        self.tab._set_status('')
        if result['returncode'] == 0:
            archive = ArchiveModel.get(name=result['params']['archive_name'])
            model = ExtractTree()
            self.tab._set_status(self.tab.tr("Processing archive contents"))
            self.tab._t = extract_dialog.ParseThread(result['data'], model)
            self.tab._t.finished.connect(lambda: self.extract_show_dialog(archive, model))
            self.tab._t.start()

    def extract_show_dialog(self, archive, model):
        """Show the dialog for choosing the archive contents to extract."""
        self.tab._set_status('')

        def process_result():
            def receive():
                extraction_folder = dialog.selectedFiles()
                if extraction_folder:
                    params = BorgExtractJob.prepare(self.tab.profile(), archive.name, model, extraction_folder[0])
                    if params['ok']:
                        self.tab._toggle_all_buttons(False)
                        job = BorgExtractJob(params['cmd'], params, self.tab.profile().repo.id)
                        job.updated.connect(self.tab.mountErrors.setText)
                        job.result.connect(self.extract_archive_result)
                        self.tab.app.jobs_manager.add_job(job)
                    else:
                        self.tab._set_status(params['message'])

            dialog = choose_file_dialog(self.tab, self.tab.tr("Choose Extraction Point"), want_folder=True)
            dialog.open(receive)

        window = ExtractDialog(archive, model)
        self.tab._toggle_all_buttons(True)
        window.setParent(self.tab, QtCore.Qt.WindowType.Sheet)
        self.tab._window = window  # for testing
        window.show()
        window.accepted.connect(process_result)

    def extract_archive_result(self, result):
        """Finished extraction."""
        self.tab._toggle_all_buttons(True)



import sys

try:
    from Cocoa import NSURL, NSBundle
    from CoreFoundation import kCFAllocatorDefault
    from Foundation import NSDictionary
    from LaunchServices import (
        LSSharedFileListCopySnapshot,
        LSSharedFileListCreate,
        LSSharedFileListInsertItemURL,
        LSSharedFileListItemRemove,
        LSSharedFileListItemResolve,
        kLSSharedFileListItemHidden,
        kLSSharedFileListItemLast,
        kLSSharedFileListNoUserInteraction,
        kLSSharedFileListSessionLoginItems,
    )

    APP_PATH = NSBundle.mainBundle().bundlePath()
except ImportError:
    pass

AUTOSTART_DELAY = """StartupNotify=false
X-GNOME-Autostart-enabled=true
X-GNOME-Autostart-Delay=20"""


def open_app_at_startup(enabled=True):
    """
    On macOS, this function adds/removes the current app bundle from Login items
    while on Linux it adds a .desktop file at ~/.config/autostart
    """
    if sys.platform == 'darwin':
        url = NSURL.alloc().initFileURLWithPath_(APP_PATH)
        login_items = LSSharedFileListCreate(kCFAllocatorDefault, kLSSharedFileListSessionLoginItems, None)
        props = NSDictionary.dictionaryWithObject_forKey_(True, kLSSharedFileListItemHidden)

        if enabled:
            LSSharedFileListInsertItemURL(login_items, kLSSharedFileListItemLast, None, None, url, props, None)
        else:
            # From https://github.com/pudquick/pyLoginItems/blob/master/pyLoginItems.py
            list_ref = LSSharedFileListCreate(None, kLSSharedFileListSessionLoginItems, None)
            login_items, _ = LSSharedFileListCopySnapshot(list_ref, None)
            flags = kLSSharedFileListNoUserInteraction + kLSSharedFileListNoUserInteraction
            for i in login_items:
                err, a_CFURL, a_FSRef = LSSharedFileListItemResolve(i, flags, None, None)
                if 'Vorta.app' in str(a_CFURL):
                    LSSharedFileListItemRemove(list_ref, i)

    elif sys.platform.startswith('linux'):
        from pathlib import Path

        from platformdirs import user_config_path

        is_flatpak = Path('/.flatpak-info').exists()

        with open(Path(__file__).parent / "assets/metadata/com.borgbase.Vorta.desktop") as desktop_file:
            desktop_file_text = desktop_file.read()

        # Find XDG_CONFIG_HOME unless when running in flatpak
        if is_flatpak:
            autostart_path = Path.home() / '.config' / 'autostart'
        else:
            autostart_path = user_config_path("autostart")

        if not autostart_path.exists():
            autostart_path.mkdir(parents=True, exist_ok=True)

        autostart_file_path = autostart_path / 'vorta.desktop'

        if enabled:
            # Replace command for flatpak if appropriate and start in background
            desktop_file_text = desktop_file_text.replace(
                "Exec=vorta",
                "Exec=flatpak run com.borgbase.Vorta --daemonize" if is_flatpak else "Exec=vorta --daemonize",
            )
            # Add autostart delay
            desktop_file_text += AUTOSTART_DELAY

            autostart_file_path.write_text(desktop_file_text)
        elif autostart_file_path.exists():
            autostart_file_path.unlink()



from pathlib import Path
from typing import List

from PyQt6 import uic
from PyQt6.QtCore import QModelIndex, QObject, Qt
from PyQt6.QtGui import QCursor, QStandardItem, QStandardItemModel
from PyQt6.QtWidgets import QAbstractItemView, QApplication, QMenu, QMessageBox, QStyledItemDelegate

from vorta.filedialog import VortaFileSelector
from vorta.i18n import trans_late, translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.store.models import ExclusionModel
from vorta.utils import get_asset
from vorta.views.utils import get_colored_icon, get_exclusion_presets

uifile = get_asset('UI/dialogs/archive/exclude.ui')
ExcludeDialogUi, ExcludeDialogBase = uic.loadUiType(uifile)


class MandatoryInputItemModel(QStandardItemModel):
    '''
    A model that prevents the user from adding an empty item to the list.
    '''

    def __init__(self, profile, parent=None):
        super().__init__(parent)
        self.profile = profile

    def setData(self, index: QModelIndex, value, role: int = ...) -> bool:
        # When a user-added item in edit mode has no text, remove it from the list.
        if role == Qt.ItemDataRole.EditRole and value == '':
            self.removeRow(index.row())
            return True
        if role == Qt.ItemDataRole.EditRole and ExclusionModel.get_or_none(name=value, profile=self.profile):
            self.removeRow(index.row())
            QMessageBox.critical(
                self.parent(),
                'Error',
                'This exclusion already exists.',
            )
            return False

        return super().setData(index, value, role)


class ExcludeDialog(ExcludeDialogBase, ExcludeDialogUi):
    def __init__(self, profile, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.profile = profile

        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.allPresets = get_exclusion_presets()
        self.buttonBox.rejected.connect(self.close)

        self.customExclusionsModel = MandatoryInputItemModel(profile=profile)
        self.customExclusionsList.setModel(self.customExclusionsModel)
        self.customExclusionsModel.itemChanged.connect(self.custom_item_changed)
        self.customExclusionsList.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.customExclusionsList.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.customExclusionsList.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.customExclusionsList.setAlternatingRowColors(True)
        self.customExclusionsListDelegate = QStyledItemDelegate()
        self.customExclusionsList.setItemDelegate(self.customExclusionsListDelegate)
        self.customExclusionsListDelegate.closeEditor.connect(self.custom_pattern_editing_finished)
        # allow removing items with the delete key with event filter
        self.installEventFilter(self)
        # context menu
        self.customExclusionsList.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customExclusionsList.customContextMenuRequested.connect(self.custom_exclusions_context_menu)

        self.exclusionPresetsModel = QStandardItemModel()
        self.exclusionPresetsList.setModel(self.exclusionPresetsModel)
        self.exclusionPresetsModel.itemChanged.connect(self.preset_item_changed)
        self.exclusionPresetsList.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.exclusionPresetsList.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.exclusionPresetsList.setAlternatingRowColors(True)

        self.exclusionsPreviewText.setReadOnly(True)

        self.rawExclusionsText.textChanged.connect(self.raw_exclusions_saved)

        self.bRemovePattern.clicked.connect(self.remove_pattern)
        self.bRemovePattern.setIcon(get_colored_icon('minus'))
        self.bPreviewCopy.clicked.connect(self.copy_preview_to_clipboard)
        self.bPreviewCopy.setIcon(get_colored_icon('copy'))
        self.bAddPattern.clicked.connect(self.add_pattern_custom_exclusion)
        self.bAddPattern.setIcon(get_colored_icon('plus'))
        self.bOpenExcludeFD.clicked.connect(self.open_exclude_filedialog)
        self.bAddPatternExcludeIfPresent.setIcon(get_colored_icon('plus'))
        self.bAddPatternExcludeIfPresent.clicked.connect(self.add_pattern_exclude_if_present)
        self.bRemovePatternExcludeIfPresent.setIcon(get_colored_icon('minus'))
        self.bRemovePatternExcludeIfPresent.clicked.connect(self.remove_pattern_exclude_if_present)

        # help text
        self.customPresetsHelpText.setOpenExternalLinks(True)
        self.custom_presets_help_text = trans_late(
            "CustomPresetsHelp",
            "Patterns that you add here will be used to exclude files and folders from the backup. "
            "For more info on how to use patterns, see the %1. "
            "To add multiple patterns at once, use the \"Raw\" tab.",
        )
        self.custom_presets_help_link_text = trans_late("CustomPresetsHelp", "documentation")
        self._set_custom_presets_help()
        self.exclusionPresetsHelpText.setText(
            translate(
                "ExclusionPresetsHelp",
                "These presets are provided by the community and are a good starting point for excluding certain types of files. You can enable or disable them as you see fit. To see the patterns that are used for each preset, switch to the \"Preview\" tab after enabling it.",  # noqa: E501
            )
        )
        self.rawExclusionsHelpText.setText(
            translate(
                "RawExclusionsHelp",
                "You can use this field to add multiple patterns at once. Each pattern should be on a separate line.",
            )
        )
        self.exclusionsPreviewHelpText.setText(
            translate(
                "ExclusionsPreviewHelp",
                "This is a preview of the patterns that will be used to exclude files and folders from the backup.",
            )
        )
        self.excludeIfPresentHelpText.setText(
            translate(
                "ExcludeIfPresentHelp",
                "Folders that contain the following files will be excluded from backups.",
            )
        )

        # Add the Exclude If Present tab
        self.excludeIfPresentModel = MandatoryInputItemModel(profile=profile)
        self.excludeIfPresentList.setModel(self.excludeIfPresentModel)
        self.excludeIfPresentList.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.excludeIfPresentList.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.excludeIfPresentList.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.excludeIfPresentList.setAlternatingRowColors(True)
        self.excludeIfPresentListDelegate = QStyledItemDelegate()
        self.excludeIfPresentList.setItemDelegate(self.excludeIfPresentListDelegate)
        self.excludeIfPresentListDelegate.closeEditor.connect(self.exclude_if_present_pattern_editing_finished)
        self.excludeIfPresentList.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.excludeIfPresentList.customContextMenuRequested.connect(self.exclude_if_present_context_menu)

        self.excludeIfPresentModel.itemChanged.connect(self.exclude_if_present_item_changed)

        self.populate_custom_exclusions_list()
        self.populate_presets_list()
        self.populate_raw_exclusions_text()
        self.populate_preview_tab()
        self.populate_exclude_if_present_patterns()

    def _set_custom_presets_help(self):
        template = self.customPresetsHelpText.text()
        sentence_template = translate(
            "CustomPresetsHelp",
            self.custom_presets_help_text,
        )
        link_text = translate("CustomPresetsHelp", self.custom_presets_help_link_text)
        sentence = format_richtext(
            escape(sentence_template),
            link(
                "https://borgbackup.readthedocs.io/en/stable/usage/help.html#borg-patterns",
                link_text,
            ),
        )
        self.customPresetsHelpText.setText(format_richtext(template, sentence))

    def populate_custom_exclusions_list(self):
        user_excluded_patterns = {
            e.name: e.enabled
            for e in self.profile.exclusions.select()
            .where(ExclusionModel.source == ExclusionModel.SourceFieldOptions.CUSTOM.value)
            .order_by(ExclusionModel.name)
        }

        for exclude, enabled in user_excluded_patterns.items():
            item = QStandardItem(exclude)
            item.setCheckable(True)
            item.setCheckState(Qt.CheckState.Checked if enabled else Qt.CheckState.Unchecked)
            self.customExclusionsModel.appendRow(item)

    def create_context_menu(self, list_widget, remove_method, toggle_method=None):
        pos = list_widget.viewport().mapFromGlobal(QCursor.pos())
        index = list_widget.indexAt(pos)
        if not index.isValid():
            return

        selected_rows = list_widget.selectedIndexes()

        if selected_rows and index not in selected_rows:
            return

        menu = QMenu(list_widget)
        menu.addAction(
            get_colored_icon('copy'),
            self.tr('Copy'),
            lambda: QApplication.clipboard().setText(index.data()),
        )

        menu.addAction(
            get_colored_icon('minus'),
            self.tr('Remove'),
            lambda: remove_method(index if not selected_rows else None),
        )

        if toggle_method:
            menu.addAction(
                get_colored_icon('check-circle'),
                self.tr('Toggle'),
                lambda: toggle_method(index if not selected_rows else None),
            )

        menu.popup(list_widget.viewport().mapToGlobal(pos))

    def custom_exclusions_context_menu(self, pos):
        self.create_context_menu(self.customExclusionsList, self.remove_pattern, self.toggle_custom_pattern)

    def populate_presets_list(self):
        for preset_slug in self.allPresets.keys():
            item = QStandardItem(self.allPresets[preset_slug]['name'])
            item.setCheckable(True)
            item.setData(preset_slug, Qt.ItemDataRole.UserRole)
            preset_model = ExclusionModel.get_or_none(
                name=preset_slug,
                source=ExclusionModel.SourceFieldOptions.PRESET.value,
                profile=self.profile,
            )

            if preset_model:
                item.setCheckState(Qt.CheckState.Checked if preset_model.enabled else Qt.CheckState.Unchecked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)

            self.exclusionPresetsModel.appendRow(item)

    def populate_raw_exclusions_text(self):
        raw_excludes = self.profile.exclude_patterns
        if raw_excludes:
            self.rawExclusionsText.setPlainText(raw_excludes)

    def populate_preview_tab(self):
        preview = self.profile.get_combined_exclusion_string()
        if self.profile.exclude_if_present:
            preview += '\n# Exclude if present'
            for f in self.profile.exclude_if_present.split('\n'):
                f = f.strip()
                if f.startswith('[x]'):
                    preview += '\n' + f[3:].strip()
        self.exclusionsPreviewText.setPlainText(preview)

    def copy_preview_to_clipboard(self):
        cb = QApplication.clipboard()
        cb.clear(mode=cb.Mode.Clipboard)
        cb.setText(self.exclusionsPreviewText.toPlainText(), mode=cb.Mode.Clipboard)

    def remove_pattern(self, index=None):
        '''
        Remove the selected item(s) from the list and the database.
        If there is no index, this was called from the context menu and the indexes are passed in.
        '''
        if not index:
            indexes = self.customExclusionsList.selectedIndexes()
            for index in reversed(sorted(indexes)):
                ExclusionModel.delete().where(
                    ExclusionModel.name == index.data(),
                    ExclusionModel.source == ExclusionModel.SourceFieldOptions.CUSTOM.value,
                    ExclusionModel.profile == self.profile,
                ).execute()
                self.customExclusionsModel.removeRow(index.row())
        else:
            ExclusionModel.delete().where(
                ExclusionModel.name == index.data(),
                ExclusionModel.source == ExclusionModel.SourceFieldOptions.CUSTOM.value,
                ExclusionModel.profile == self.profile,
            ).execute()
            self.customExclusionsModel.removeRow(index.row())

        self.populate_preview_tab()

    def toggle_custom_pattern(self, index=None):
        '''
        Toggle the check state of the selected item(s).
        If there is no index, this was called from the context menu and the indexes are passed in.
        '''
        if not index:
            indexes = self.customExclusionsList.selectedIndexes()
            for index in indexes:
                item = self.customExclusionsModel.itemFromIndex(index)
                if item.checkState() == Qt.CheckState.Checked:
                    item.setCheckState(Qt.CheckState.Unchecked)
                else:
                    item.setCheckState(Qt.CheckState.Checked)
        else:
            item = self.customExclusionsModel.itemFromIndex(index)
            if item.checkState() == Qt.CheckState.Checked:
                item.setCheckState(Qt.CheckState.Unchecked)
            else:
                item.setCheckState(Qt.CheckState.Checked)

    def add_pattern(self, list_widget, model):
        '''
        Add an empty item to the list in editable mode.
        Don't add an item if the user is already editing an item.
        '''
        if list_widget.state() == QAbstractItemView.State.EditingState:
            return
        item = QStandardItem('')
        item.setCheckable(True)
        item.setCheckState(Qt.CheckState.Checked)
        model.appendRow(item)
        list_widget.edit(item.index())
        list_widget.scrollToBottom()

    def add_pattern_custom_exclusion(self):
        self.add_pattern(self.customExclusionsList, self.customExclusionsModel)

    def text_editing_finished(self, model):
        '''
        Go through all items in the list and if any of them are empty, remove them.
        Handles the case where the user presses the escape key to cancel editing.
        '''
        for row in range(model.rowCount()):
            item = model.item(row)
            if item is not None and item.text().strip() == '':
                model.removeRow(row)

    def custom_pattern_editing_finished(self, editor):
        self.text_editing_finished(self.customExclusionsModel)

    def custom_item_changed(self, item):
        '''
        When the user checks or unchecks an item, update the database.
        When the user adds a new item, add it to the database.
        '''
        if item.text().strip():  # Ensure the item text is not empty before adding to DB
            if not ExclusionModel.get_or_none(
                name=item.text(), source=ExclusionModel.SourceFieldOptions.CUSTOM.value, profile=self.profile
            ):
                ExclusionModel.create(
                    name=item.text(), source=ExclusionModel.SourceFieldOptions.CUSTOM.value, profile=self.profile
                )

            ExclusionModel.update(enabled=item.checkState() == Qt.CheckState.Checked).where(
                ExclusionModel.name == item.text(),
                ExclusionModel.source == ExclusionModel.SourceFieldOptions.CUSTOM.value,
                ExclusionModel.profile == self.profile,
            ).execute()

        self.populate_preview_tab()

    def preset_item_changed(self, item):
        '''
        Create or update the preset in the database.
        If the user unchecks the preset, set enabled to False, otherwise set it to True.
        If the preset doesn't exist, create it and set enabled to True.
        '''
        preset = ExclusionModel.get_or_none(
            name=item.data(Qt.ItemDataRole.UserRole),
            source=ExclusionModel.SourceFieldOptions.PRESET.value,
            profile=self.profile,
        )
        if preset:
            preset.enabled = item.checkState() == Qt.CheckState.Checked
            preset.save()
        else:
            ExclusionModel.create(
                name=item.data(Qt.ItemDataRole.UserRole),
                source=ExclusionModel.SourceFieldOptions.PRESET.value,
                profile=self.profile,
                enabled=item.checkState() == Qt.CheckState.Checked,
            )

        self.populate_preview_tab()

    def exclude_if_present_item_changed(self, item):
        self.save_exclude_if_present_patterns()

    def exclude_if_present_context_menu(self, pos):
        self.create_context_menu(self.excludeIfPresentList, self.remove_pattern_exclude_if_present)

    def remove_pattern_exclude_if_present(self, index=None):
        if not index:
            indexes = self.excludeIfPresentList.selectedIndexes()
            for index in reversed(sorted(indexes)):
                self.excludeIfPresentModel.removeRow(index.row())
        else:
            self.excludeIfPresentModel.removeRow(index.row())
        self.save_exclude_if_present_patterns()

    def add_pattern_exclude_if_present(self):
        self.add_pattern(self.excludeIfPresentList, self.excludeIfPresentModel)
        self.save_exclude_if_present_patterns()

    def exclude_if_present_pattern_editing_finished(self, editor):
        self.text_editing_finished(self.excludeIfPresentModel)
        self.save_exclude_if_present_patterns()

    def save_exclude_if_present_patterns(self):
        '''
        Save Exclude If Present files to profile.
        '''
        patterns = []
        for row in range(self.excludeIfPresentModel.rowCount()):
            item = self.excludeIfPresentModel.item(row)
            text = item.text().strip()

            # Remove any existing prefix
            if text.startswith('[x] '):
                text = text[4:]
            elif text.startswith('[] '):
                text = text[3:]

            # Add the correct prefix based on the check state
            prefix = '[x] ' if item.checkState() == Qt.CheckState.Checked else '[] '
            patterns.append(prefix + text)

        self.profile.exclude_if_present = '\n'.join(patterns)
        self.profile.save()
        self.populate_preview_tab()

    def normalize_exclude_if_present_patterns(self):
        '''
        Ensure all patterns in exclude_if_present start with either [x] or [].
        '''
        patterns = self.profile.exclude_if_present.split('\n') if self.profile.exclude_if_present else []
        normalized_patterns = []
        for pattern in patterns:
            if not pattern.startswith('[x]') and not pattern.startswith('[]'):
                pattern = '[x] ' + pattern
            normalized_patterns.append(pattern)
        self.profile.exclude_if_present = '\n'.join(normalized_patterns)
        self.profile.save()

    def populate_exclude_if_present_patterns(self):
        '''
        Populate the 'Exclude If Present' list from the profile.
        '''
        self.normalize_exclude_if_present_patterns()
        patterns = self.profile.exclude_if_present.split('\n') if self.profile.exclude_if_present else []
        for pattern in patterns:
            item = QStandardItem()
            if pattern.startswith('[x]'):
                item.setText(pattern[4:])
                item.setCheckable(True)
                item.setCheckState(Qt.CheckState.Checked)
            elif pattern.startswith('[]'):
                item.setText(pattern[3:])
                item.setCheckable(True)
                item.setCheckState(Qt.CheckState.Unchecked)
            self.excludeIfPresentModel.appendRow(item)

    def raw_exclusions_saved(self):
        '''
        When the user saves changes in the raw exclusions text box, add it to the database.
        '''
        raw_excludes = self.rawExclusionsText.toPlainText()
        self.profile.exclude_patterns = raw_excludes
        self.profile.save()

        self.populate_preview_tab()

    def eventFilter(self, source, event):
        '''
        When the user presses the delete key, remove the selected items.
        '''
        if event.type() == event.Type.KeyPress and event.key() == Qt.Key.Key_Delete:
            self.remove_pattern()
            return True
        return QObject.eventFilter(self, source, event)

    def make_exclude_pattern_from_path(self, paths: List[str]) -> List[str]:
        '''
        Create exclusion patterns from a list of given paths.
        '''
        patterns = []
        # Both files and folders use the same pattern format (fm:*/name),
        # confirmed through testing to correctly exclude them during backup creation.
        for path in paths:
            patterns.append(f"pf:{path}")
        return patterns

    def open_exclude_filedialog(self):
        file_dialog = VortaFileSelector(
            self, window_title='Exclude Files and Folders', title='Select files and folders to exclude:'
        )
        paths = file_dialog.get_paths()  # Selected paths from file dialog
        if paths:
            patterns = self.make_exclude_pattern_from_path(paths)
            for pattern in patterns:
                # Check if the path already exists in the model
                if not ExclusionModel.get_or_none(name=pattern, profile=self.profile):
                    # Create a new item in the list
                    item = QStandardItem(pattern)
                    item.setCheckable(True)
                    item.setCheckState(Qt.CheckState.Checked)
                    self.customExclusionsModel.appendRow(item)

                    # Add the item to database
                    ExclusionModel.create(
                        name=pattern, source=ExclusionModel.SourceFieldOptions.CUSTOM.value, profile=self.profile
                    )
            self.populate_preview_tab()



import logging

from PyQt6.QtWidgets import QTableWidgetItem

from vorta.borg.mount import BorgMountJob
from vorta.borg.umount import BorgUmountJob
from vorta.i18n import translate
from vorta.utils import choose_file_dialog
from vorta.views.utils import get_colored_icon

logger = logging.getLogger(__name__)


class ArchiveMount:
    """Component handling the mount/unmount operations for the ArchiveTab."""

    def __init__(self, tab):
        self.tab = tab

    def bmountarchive_clicked(self):
        """
        Handle MountArchive being clicked.
        Mount or umount the current archive depending on its current state.
        """
        archive_name = self.tab.selected_archive_name()

        if not archive_name:
            logger.warning("Archive name of selection is empty.")
            return

        if archive_name in self.tab.mount_points:
            self.unmount_action(archive_name=archive_name)
        else:
            self.mount_action(archive_name=archive_name)

    def bmountrepo_clicked(self):
        """
        Handle MountRepo being clicked.
        Mount or umount the repository depending on its current state.
        """
        if self.tab.repo_mount_point:
            self.unmount_action()
        else:
            self.mount_action()

    def bmountarchive_refresh(self, icon_only=False):
        """
        Update label, tooltip and state of MountArchive.
        The new state depends on the mount status of the current archive.
        This also updates the icon of the button.
        """
        archive_name = self.tab.selected_archive_name()

        if archive_name in self.tab.mount_points:
            self.tab.bMountArchive.setIcon(get_colored_icon('eject'))
            if not icon_only:
                self.tab.bMountArchive.setText(self.tab.tr("Unmount"))
                self.tab.bMountArchive.setToolTip(self.tab.tr('Unmount the selected archive from the file system'))
        else:
            self.tab.bMountArchive.setIcon(get_colored_icon('folder-open'))
            if not icon_only:
                self.tab.bMountArchive.setText(self.tab.tr("Mount…"))
                self.tab.bMountArchive.setToolTip(
                    self.tab.tr("Mount the selected archive as a folder in the file system")
                )

    def bmountrepo_refresh(self):
        """
        Update label, tooltip and state of MountRepo.
        The new state depends on the mount status of the current archive.
        This also updates the icon of the button.
        """
        if self.tab.repo_mount_point:
            self.tab.bMountRepo.setText(self.tab.tr("Unmount"))
            self.tab.bMountRepo.setToolTip(self.tab.tr('Unmount the repository from the file system'))
            self.tab.bMountRepo.setIcon(get_colored_icon('eject'))
        else:
            self.tab.bMountRepo.setText(self.tab.tr("Mount…"))
            self.tab.bMountRepo.setIcon(get_colored_icon('folder-open'))
            self.tab.bMountRepo.setToolTip(self.tab.tr("Mount the repository as a folder in the file system"))

    def mount_action(self, archive_name=None):
        """
        Mount an archive or the whole repository.
        Opens a file chooser to let the user choose a mount point and starts
        the borg job for mounting afterwards.
        Parameters
        ----------
        archive_name : str, optional
            The archive to mount or None, by default None
        """
        profile = self.tab.profile()
        params = BorgMountJob.prepare(profile, archive=archive_name)
        if not params['ok']:
            self.tab._set_status(params['message'])
            return

        def receive():
            mount_point = dialog.selectedFiles()
            if mount_point:
                params['cmd'].append(mount_point[0])
                params['mount_point'] = mount_point[0]

                if params['ok']:
                    self.tab._toggle_all_buttons(False)
                    job = BorgMountJob(params['cmd'], params, self.tab.profile().repo.id)
                    job.updated.connect(self.tab.mountErrors.setText)
                    job.result.connect(self.mount_result)
                    self.tab.app.jobs_manager.add_job(job)

        dialog = choose_file_dialog(self.tab, self.tab.tr("Choose Mount Point"), want_folder=True)
        dialog.open(receive)

    def mount_result(self, result):
        if result['returncode'] == 0:
            self.tab._set_status(self.tab.tr('Mounted successfully.'))

            mount_point = result['params']['mount_point']

            if result['params'].get('mounted_archive'):
                # archive was mounted
                archive_name = result['params']['mounted_archive']
                self.tab.mount_points[archive_name] = mount_point

                # update column in table
                row = self.tab.row_of_archive(archive_name)
                item = QTableWidgetItem(result['cmd'][-1])
                self.tab.archiveTable.setItem(row, 3, item)

                # update button
                self.bmountarchive_refresh()
            else:
                # whole repo was mounted
                self.tab.repo_mount_point = mount_point
                self.bmountrepo_refresh()

        self.tab._toggle_all_buttons(True)

    def unmount_action(self, archive_name=None):
        """
        Unmount a (mounted) repository or archive.
        If the target isn't mounted nothing happens.
        Parameters
        ----------
        archive_name : str, optional
            The archive to unmount, by default None
        """
        if archive_name:
            # unmount a single archive
            mount_point = self.tab.mount_points.get(archive_name)
        else:
            # unmount the whole repository
            mount_point = self.tab.repo_mount_point

        if mount_point is not None:
            profile = self.tab.profile()
            params = BorgUmountJob.prepare(profile, mount_point, archive_name=archive_name)
            if not params['ok']:
                self.tab._set_status(translate('message', params['message']))
                return

            job = BorgUmountJob(params['cmd'], params, self.tab.profile().repo.id)
            job.updated.connect(self.tab.mountErrors.setText)
            job.result.connect(self.umount_result)
            self.tab.app.jobs_manager.add_job(job)

    def umount_result(self, result):
        self.tab._toggle_all_buttons(True)
        archive_name = result['params'].get('current_archive')
        mount_point = result['params']['mount_point']

        if result['returncode'] == 0:
            self.tab._set_status(self.tab.tr('Un-mounted successfully.'))

            if archive_name:
                # unmount single archive
                del self.tab.mount_points[archive_name]
                row = self.tab.row_of_archive(archive_name)
                item = QTableWidgetItem('')
                self.tab.archiveTable.setItem(row, 3, item)

                # update button
                self.bmountarchive_refresh()
            else:
                # unmount repo
                self.tab.repo_mount_point = None
                self.bmountrepo_refresh()
        else:
            self.tab._set_status(
                self.tab.tr('Unmounting failed. Make sure no programs are using {}').format(mount_point)
            )



from vorta.store.models import ArchiveModel, RepoModel
from vorta.utils import borg_compat

from .borg_job import BorgJob


class BorgInfoArchiveJob(BorgJob):
    def started_event(self):
        self.app.backup_started_event.emit()
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Refreshing archive…')}")

    def finished_event(self, result):
        self.app.backup_finished_event.emit(result)
        self.result.emit(result)
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Refreshing archive done.')}")

    @classmethod
    def prepare(cls, profile, archive_name):
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret

        ret['ok'] = True
        ret['cmd'] = ['borg', 'info', '--log-json', '--json']
        if borg_compat.check('V2'):
            ret['cmd'].extend(["-r", profile.repo.url, '-a', archive_name])
        else:
            ret['cmd'].append(f'{profile.repo.url}::{archive_name}')
        ret['archive_name'] = archive_name

        return ret

    def process_result(self, result):
        if result['returncode'] == 0:
            remote_archives = result['data'].get('archives', [])

            # get info stored during BorgJob.prepare()
            # repo_id = self.params['repo_id']
            repo_id = result['params']['repo_id']

            # Update remote archives.
            for remote_archive in remote_archives:
                archive = ArchiveModel.get_or_none(snapshot_id=remote_archive['id'], repo=repo_id)
                if archive is None:
                    # archive id was changed during rename, so we need to find it by name
                    archive = ArchiveModel.get_or_none(name=remote_archive['name'], repo=repo_id)
                    archive.snapshot_id = remote_archive['id']

                archive.name = remote_archive['name']  # in case name changed
                # archive.time = parser.parse(remote_archive['time'])
                archive.duration = remote_archive['duration']
                archive.size = remote_archive['stats']['deduplicated_size']

                archive.save()

            if 'cache' in result['data']:
                stats = result['data']['cache']['stats']
                repo = RepoModel.get(id=result['params']['repo_id'])
                repo.total_size = stats['total_size']
                repo.unique_size = stats['unique_size']
                repo.total_unique_chunks = stats['total_unique_chunks']
                repo.save()



import logging
from datetime import datetime

from PyQt6 import QtCore, uic

from vorta import config
from vorta._version import __version__
from vorta.i18n import trans_late, translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.utils import get_asset
from vorta.views.base_tab import BaseTab
from vorta.views.utils import get_colored_icon

uifile = get_asset('UI/about_tab.ui')
AboutTabUI, AboutTabBase = uic.loadUiType(uifile)

logger = logging.getLogger(__name__)


class AboutTab(BaseTab, AboutTabBase, AboutTabUI):
    refresh_archive = QtCore.pyqtSignal()

    def __init__(self, parent=None, profile_provider=None):
        """Init."""
        super().__init__(parent=parent, profile_provider=profile_provider)
        self.setupUi(parent)
        self.versionLabel.setText(__version__)
        self.bug_report_text = trans_late('Form', '%1 to report a bug.')
        self.docs_text = trans_late('Form', '%1 to view the docs.')
        self.repo_text = trans_late('Form', '%1 to view Git repo.')
        self.click_here = trans_late('Form', 'Click here')
        self.view_logs_text = trans_late('Form', 'View the logs')
        self._set_links()
        self.gpl_logo.setPixmap(get_colored_icon('gpl_logo', scaled_height=40, return_qpixmap=True))
        self.python_logo.setPixmap(get_colored_icon('python_logo', scaled_height=40, return_qpixmap=True))
        copyright_text = self.copyrightLabel.text()
        copyright_text = copyright_text.replace('2020', str(datetime.now().year))
        self.copyrightLabel.setText(copyright_text)

    def _set_links(self):
        click_here = translate('Form', self.click_here)

        bug_template = self.bugReportLink.text()
        bug_sentence = format_richtext(
            escape(translate('Form', self.bug_report_text)),
            link('https://github.com/borgbase/vorta/issues/new/choose', click_here),
        )
        self.bugReportLink.setText(format_richtext(bug_template, bug_sentence))

        docs_template = self.docsLink.text()
        docs_sentence = format_richtext(
            escape(translate('Form', self.docs_text)),
            link('https://borgbackup.readthedocs.io/en/master/index.html', click_here),
        )
        self.docsLink.setText(format_richtext(docs_template, docs_sentence))

        repo_template = self.repoLink.text()
        repo_sentence = format_richtext(
            escape(translate('Form', self.repo_text)),
            link('https://github.com/borgbase/vorta', click_here),
        )
        self.repoLink.setText(format_richtext(repo_template, repo_sentence))

        log_template = self.logLink.text()
        log_link = link(f"file://{config.LOG_DIR}", translate('Form', self.view_logs_text))
        self.logLink.setText(format_richtext(log_template, log_link))

    def set_borg_details(self, version, path):
        self.borgVersion.setText(version)
        self.borgPath.setText(f"<center>Path to Borg: {path}</center>")



from __future__ import annotations

import os
import shutil
from datetime import datetime, timedelta

import peewee as pw
from peewee import Tuple, fn
from playhouse import signals

from vorta import config
from vorta.autostart import open_app_at_startup

from .migrations import run_migrations
from .models import (
    DB,
    ArchiveModel,
    BackupProfileModel,
    EventLogModel,
    ExclusionModel,
    RepoModel,
    RepoPassword,
    SchemaVersion,
    SettingsModel,
    SourceFileModel,
    WifiSettingModel,
)
from .settings import get_misc_settings

SCHEMA_VERSION = 23


@signals.post_save(sender=SettingsModel)
def setup_autostart(model_class: type, instance: SettingsModel, created: bool) -> None:
    if instance.key == 'autostart':
        open_app_at_startup(instance.value)


def cleanup_db() -> None:
    # Clean up database
    DB.execute_sql("VACUUM")
    DB.close()


def init_db(con: pw.SqliteDatabase | None = None) -> None:
    if con is not None:
        os.umask(0o0077)
        DB.initialize(con)
        DB.connect()
    DB.create_tables(
        [
            RepoModel,
            RepoPassword,
            BackupProfileModel,
            SourceFileModel,
            SettingsModel,
            ArchiveModel,
            WifiSettingModel,
            EventLogModel,
            SchemaVersion,
            ExclusionModel,
        ]
    )

    # Delete old log entries after 6 months.
    # The last `create` command of each profile must not be deleted
    # since the scheduler uses it to determine the last backup time.
    last_backups_per_profile = (
        EventLogModel.select(EventLogModel.profile, fn.MAX(EventLogModel.start_time))
        .where(EventLogModel.subcommand == 'create')
        .group_by(EventLogModel.profile)
    )
    last_scheduled_backups_per_profile = (
        EventLogModel.select(EventLogModel.profile, fn.MAX(EventLogModel.start_time))
        .where(EventLogModel.subcommand == 'create', EventLogModel.category == 'scheduled')
        .group_by(EventLogModel.profile)
    )

    six_months_ago = datetime.now() - timedelta(days=6 * 30)
    entry = Tuple(EventLogModel.profile, EventLogModel.start_time)
    EventLogModel.delete().where(
        EventLogModel.start_time < six_months_ago,
        entry.not_in(last_backups_per_profile),
        entry.not_in(last_scheduled_backups_per_profile),
    ).execute()

    # Migrations
    current_schema, created = SchemaVersion.get_or_create(id=1, defaults={'version': SCHEMA_VERSION})
    current_schema.save()
    if created or current_schema.version == SCHEMA_VERSION:
        pass
    elif con is not None:
        backup_current_db(current_schema.version)
        run_migrations(current_schema, con)

    # Create missing settings and update labels.
    # Leave only setting values untouched.
    for setting in get_misc_settings():
        s, created = SettingsModel.get_or_create(key=setting['key'], defaults=setting)
        s.label = setting['label']
        s.type = setting['type']

        if 'group' in setting:
            s.group = setting['group']
        if 'tooltip' in setting:
            s.tooltip = setting['tooltip']

        s.save()


def backup_current_db(schema_version: int) -> None:
    """
    Creates a backup copy of settings.db
    """

    assert config.SETTINGS_DIR is not None
    timestamp = datetime.now().strftime('%Y-%m-%d-%H%M%S')
    backup_file_name = f'settings_v{schema_version}_{timestamp}.db'
    shutil.copy(config.SETTINGS_DIR / 'settings.db', config.SETTINGS_DIR / backup_file_name)



"""
Implementation of a tree model for use with `QTreeView` based on (file) paths.

"""

import bisect
import enum
import os.path as osp
from functools import reduce
from pathlib import PurePath
from typing import Generic, List, Optional, Sequence, Tuple, TypeVar, Union, overload

from PyQt6.QtCore import (
    QAbstractItemModel,
    QModelIndex,
    QObject,
    QSortFilterProxyModel,
    Qt,
    pyqtSignal,
)

#: A representation of a path
Path = Tuple[str, ...]
PathLike = Union[Path, Sequence[str]]


def relative_path(p1: PathLike, p2: PathLike) -> Path:
    """Get p2 relative to p1."""
    if len(p2) <= len(p1):
        return ()

    return tuple(p2[len(p1) :])


def path_to_str(path: PathLike) -> str:
    """Return a string representation of a path."""
    if not path:
        return ''

    return osp.join(*path)


#: Type of FileSystemItem's data
T = TypeVar('T')
FileSystemItemLike = Union[Tuple[Union[PurePath, Path], Optional[T]], 'FileSystemItem']

#: Default return value
A = TypeVar('A')


class FileSystemItem(Generic[T]):
    """
    An item in the virtual file system.

    ..warning::

        Do not edit `children` manually. Always use `add` or `remove` or
        `sort`.

    Attributes
    ----------
    path : Path
        The path of this item.
    data : Any
        The data belonging to this item.
    children : List[FileSystemItem]
        The children of this item.
    _subpath : str
        The subpath of this item relative to its parent.
    _parent : FileSystemItem or None
        The parent of the item.
    """

    __slots__ = ['path', 'children', 'data', '_parent', 'subpath']

    def __init__(self, path: PathLike, data: T):
        """Init."""
        self.path: Path = tuple(path)
        self.data = data
        self.subpath: str = None
        self.children: List[FileSystemItem[T]] = []
        self._parent: Optional[FileSystemItem[T]] = None

    # @property
    # def subpath(self) -> str:
    #     """
    #     Get the name of the item which is the subpath relative to its parent.
    #     """
    #     return self.path[-1]

    # @property
    # def children(self):
    #     """Get an iterable view of the item's children."""
    #     return self.child_map.values()

    def add(self, child: 'FileSystemItem[T]', _subpath: str = None, _check: bool = True):
        """
        Add a child.

        The parameters starting with an underscore exist for performance
        reasons only. They should only be used if the operations that these
        parameters toggle were performed already.

        Parameters
        ----------
        child : FileSystemItem
            The child to add.
        _subpath : str, optional
            Precalculated subpath, default is None.
        _check : bool, optional
            Whether to check for children with the same subpath (using `get`).
        """
        if _subpath is not None:
            child.subpath = _subpath
        else:
            child.subpath = path_to_str(relative_path(self.path, child.path))

        i = bisect.bisect(self.children, child)

        # check for a child with the same subpath
        if _check and len(self.children) > i - 1 >= 0 and self.children[i - 1].subpath == child.subpath:
            raise RuntimeError("The subpath must be unique to a parent's children.")

        # add
        child._parent = self
        self.children.insert(i, child)

    def addChildren(self, children: List['FileSystemItem[T]']):
        """
        Add a list of children.

        Parameters
        ----------
        children : List[FileSystemItem]
            The children to add.
        """
        for child in children:
            self.add(child)

    @overload
    def remove(self, subpath: str) -> None:
        pass

    @overload
    def remove(self, index: int) -> None:
        pass

    @overload
    def remove(self, child: 'FileSystemItem[T]') -> None:
        pass

    def remove(self, child_subpath_index):
        """
        Remove the given children.

        The index or child to remove must be in the list
        else an error will be raised.

        Parameters
        ----------
        child_or_index : FileSystemItem or int
            The instance to remove or its index in `children`.

        Raises
        ------
        ValueError
            The given item is not a child of this one.
        IndexError
            The given index is not a valid one.
        """
        if isinstance(child_subpath_index, FileSystemItem):
            child = child_subpath_index
            if not child.subpath:
                raise ValueError("Child without subpath")

            i = bisect.bisect_left(self.children, child)
            if i < len(self.children) and self.children[i] == child:
                del self.children[i]
            else:
                raise ValueError("Child not found")

        elif isinstance(child_subpath_index, str):
            subpath = child_subpath_index
            i = bisect.bisect_left(self.children, subpath)
            if i < len(self.children) and self.children[i].subpath == subpath:
                del self.children[i]
            else:
                raise ValueError("Child not found")

        elif isinstance(child_subpath_index, int):
            i = child_subpath_index
            del self.children[i]

        else:
            raise TypeError(
                "First argument passed to `{}.remove` has invalid type {}".format(
                    type(self).__name__, type(child_subpath_index).__name__
                )
            )

    def __getitem__(self, index: int):
        """
        Get a an item.

        This allows accessing the attributes in the same manner for instances
        of this type and instances of `FileSystemItemLike`.
        """
        if index == 0:
            return self.path
        elif index == 1:
            return self.data
        else:
            raise IndexError("Index {} out of range(0, 2)".format(index))

    def get(self, subpath: str, default: Optional[A] = None) -> Union[Tuple[int, 'FileSystemItem[T]'], Optional[A]]:
        """
        Find direct child with given subpath.

        Parameters
        ----------
        subpath : str
            The items subpath relative to this.
        default : Any, optional
            The item to return if the child wasn't found, default None.

        Returns
        -------
        Tuple[int, FileSystemItem] or None
            The index and item if found else `default`.
        """
        i = bisect.bisect_left(self.children, subpath)
        if i < len(self.children):
            child = self.children[i]
            if child.subpath == subpath:
                return i, child
        return default

    def get_path(self, path: PathLike) -> Optional['FileSystemItem[T]']:
        """
        Get the item with the given subpath relative to this item.

        Parameters
        ----------
        path : Path
            The subpath.
        """

        def walk(fsi, pp):
            if fsi is None:
                return None
            res = fsi.get(pp)
            if not res:
                return None
            i, item = res
            return item

        fsi = reduce(walk, path, self)  # handles empty path -> returns self
        return fsi

    def __repr__(self):
        """Get a string representation."""
        return "FileSystemItem<'{}', '{}', {}, {}>".format(
            self.path,
            self.subpath,
            self.data,
            [c.subpath for c in self.children],
        )

    def __lt__(self, other):
        """Lower than for bisect sorting."""
        if isinstance(other, FileSystemItem):
            return self.subpath < other.subpath
        if isinstance(other, (list, tuple)):
            for s, o in zip(self.path, other):
                if s != o:
                    return s < o
            else:
                return len(self.path) < len(other)
        else:
            return self.subpath < other

    def __gt__(self, other):
        """Greater than for bisect sorting."""
        if isinstance(other, FileSystemItem):
            return self.subpath > other.subpath
        if isinstance(other, (list, tuple)):
            for s, o in zip(self.path, other):
                if s != o:
                    return s > o
            else:
                return len(self.path) > len(other)
        else:
            return self.subpath > other


class FileTreeModel(QAbstractItemModel, Generic[T]):
    """
    FileTreeModel managing a virtual file system with variable file data.

    Attributes
    ----------
    mode: DisplayMode
        The tree display mode of the model.

    Methods
    -------
    _make_filesystemitem(path, data, children)
        Construct a `FileSystemItem`.
    _merge_data(item, data)
        Add the given data to the item.
    _flat_filter
        Return whether an item is part of the flat model representation.
    flags
    columnCount
    headerData

    """

    class DisplayMode(enum.Enum):
        """
        The tree display modes available for the model.

        """

        #: normal file tree
        TREE = enum.auto()

        #: combine items in the tree having a single child with that child
        SIMPLIFIED_TREE = enum.auto()

        #: simple list of items
        FLAT = enum.auto()

    def __init__(self, mode: 'FileTreeModel.DisplayMode' = DisplayMode.TREE, parent=None):
        """Init."""
        super().__init__(parent)
        self.root: FileSystemItem[T] = FileSystemItem([], None)

        self.mode = mode
        #: flat representation of the tree
        self._flattened: List[FileSystemItem] = []

    def addItems(self, items: List[FileSystemItemLike[T]]):
        """
        Add file system items to the model (batched).

        This method wraps all additions in a single model reset,
        making it much more efficient than calling `addItem` repeatedly.

        Parameters
        ----------
        items : List[FileSystemItemLike]
            The items.
        """
        if not items:
            return

        self.beginResetModel()
        for item in items:
            self._addItemInternal(item)
        self.endResetModel()

    def addItem(self, item: FileSystemItemLike[T]):
        """
        Add a file system item to the model.

        For adding multiple items, use `addItems()` instead as it
        batches the model reset for better performance.

        Parameters
        ----------
        item : FileSystemItemLike
            The item.
        """
        self.beginResetModel()
        self._addItemInternal(item)
        self.endResetModel()

    def _addItemInternal(self, item: FileSystemItemLike[T]):
        """
        Internal: add item without signaling model reset.

        Parameters
        ----------
        item : FileSystemItemLike
            The item.
        """
        path = item[0]
        data = item[1]

        if isinstance(path, PurePath):
            path = path.parts

        if not path:
            return  # empty path (e.g. `.`) can't be added

        def child(tup, subpath):
            fsi, i = tup
            i += 1
            return self._addChild(fsi, path[:i], subpath, None), i

        fsi, dummy = reduce(child, path[:-1], (self.root, 0))

        self._addChild(fsi, path, path[-1], data)

    def _addChild(
        self, item: FileSystemItem[T], path: PathLike, path_part: str, data: Optional[T]
    ) -> FileSystemItem[T]:
        """
        Add a child to an item.

        This is called by `addItem` in a reduce statement. It should
        add a new child with the given attributes to the given item.
        This implementation provides a reasonable default, most subclasses
        won't need to override this method. The implementation should make use
        of `_make_filesystemitem`, `_merge_data`, `_add_children`.

        Parameters
        ----------
        item : FileSystemItem
            The item to add a new child to.
        path : PathLike
            The path of the new child.
        path_part : str
            The subpath of the new child relative to `item`.
        data : Any or None
            The data of the new child.
        children : Any or None
            The initial children of the item.

        Returns
        -------
        FileSystemItem
            [description]
        """
        res = item.get(path_part)
        if res:
            i, child = res
            if data is not None:
                self._merge_data(child, data)
        else:
            child = self._make_filesystemitem(path, data)

            if self._flat_filter(child):
                i = bisect.bisect(self._flattened, child.path)
                self._flattened.insert(i, child)

            item.add(child, _subpath=path_part, _check=False)

            # update parent data
            self._process_child(child)

        return child

    def _make_filesystemitem(self, path: PathLike, data: Optional[T]) -> FileSystemItem[T]:
        """
        Construct a `FileSystemItem`.

        The attributes are the attributes of a `FileSystemItemLike`.
        This implementation already provides reasonable default that
        subclasses can be used.

        ..warning:: Do always call `_addChild` to add a child to an item.

        Parameters
        ----------
        path : PathLike
            The path of the item.
        data : Any or None
            The data.
        children : Any or None
            The initial children.

        Returns
        -------
        FileSystemItem
            The FileSystemItem for the internal tree structure.
        """
        return FileSystemItem(path, data)

    def _process_child(self, child: FileSystemItem[T]):
        """
        Process a new child.

        This can make some changes to the child's data like
        setting a default value if the child's data is None.
        This can also update the data of the parent.

        Parameters
        ----------
        child : FileSystemItem
            The child that was added.
        """
        pass  # Does nothing

    def _merge_data(self, item: FileSystemItem[T], data: Optional[T]):
        """
        Add the given data to the item.

        This method is called by `_addChild` which in turn is called by
        `addItem`. It gets an item in the virtual file system that was
        added again with the given data. The data may be None.

        Parameters
        ----------
        item : FileSystemItem
            The item to merge the data in.
        data : Any or None
            The data to add.
        """
        if not item.data:
            item.data = data

    def removeItem(self, path: Union[PurePath, PathLike]) -> None:
        """
        Remove an item from the model.

        Parameters
        ----------
        path : PathLike or PurePath
            The path of the item to remove.
        """
        if isinstance(path, PurePath):
            path = path.parts

        if not path:
            return

        self.beginResetModel()

        parent = self.getItem(path[:-1])

        if not parent:
            return

        res = parent.get(path[-1])

        if not res:
            return

        i, item = res

        # remove item and its children in flat representation
        items_to_remove: List[FileSystemItem] = [item]
        while items_to_remove:
            to_remove = items_to_remove.pop()

            fi = bisect.bisect_left(self._flattened, to_remove.path)
            if fi < len(self._flattened) and self._flattened[fi] is to_remove:
                del self._flattened[fi]

            items_to_remove.extend(to_remove.children)

        # remove item from tree representation
        parent.remove(i)

        self.endResetModel()

    def setMode(self, value: 'DisplayMode'):
        """
        Set the display mode of the tree model.

        In TREE mode (default) the tree will be displayed as is.
        In SIMPLIFIED_TREE items will simplify the tree by combining
        items with their single child if they possess only one.
        In FLAT mode items will be displayed as a simple list. The items
        shown can be filtered by `_flat_filter`.

        Parameters
        ----------
        value : bool
            The new value for the attribute.

        See also
        --------
        getMode: Get the current mode.
        _flat_filter
        """
        if value == self.mode:
            return  # nothing to do

        self.beginResetModel()
        self.mode = value
        self.endResetModel()

    def getMode(self) -> bool:
        """
        Get the display mode set.

        Returns
        -------
        DisplayMode
            The current value.

        See also
        --------
        setMode : Set the mode.
        """
        return self.mode

    def _flat_filter(self, item: FileSystemItem[T]) -> bool:
        """
        Return whether an item is part of the flat model representation.
        """
        return True

    def _simplify_filter(self, item: FileSystemItem[T]) -> bool:
        """
        Return whether an item may be merged in simplified mode.
        """
        return True

    def getItem(self, path: Union[PurePath, PathLike]) -> Optional[FileSystemItem[T]]:
        """
        Get the item with the given path.

        Parameters
        ----------
        path : PathLike or PurePath
            The path of the item.

        Returns
        -------
        Optional[FileSystemItem]
            [description]
        """
        if isinstance(path, PurePath):
            path = path.parts

        return self.root.get_path(path)  # handles empty path

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        """
        Get the data for the given role and index.

        The indexes internal pointer references the corresponding
        `FileSystemItem`.

        Parameters
        ----------
        index : QModelIndex
            The index of the item.
        role : int, optional
            The data role, by default Qt.ItemDataRole.DisplayRole

        Returns
        -------
        Any
            The data, return None if no data is available for the role.
        """
        return super().data(index, role)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        """
        Returns the number of rows under the given parent.

        When the parent is valid it means that rowCount is returning
        the number of children of parent.

        Parameters
        ----------
        parent : QModelIndex, optional
            The index of the parent item, by default QModelIndex()

        Returns
        -------
        int
            The number of children.
        """
        if parent.column() > 0:
            return 0  # Only the first column has children

        # flat mode
        if self.mode == self.DisplayMode.FLAT:
            if not parent.isValid():
                return len(self._flattened)
            return 0

        # tree mode
        if not parent.isValid():
            parent_item: FileSystemItem = self.root
        else:
            parent_item = parent.internalPointer()

        return len(parent_item.children)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        """
        Returns the number of columns for the children of the given parent.

        This corresponds to the number of data (column) entries shown
        for each item in the tree view.

        Parameters
        ----------
        parent : QModelIndex, optional
            The index of the parent, by default QModelIndex()

        Returns
        -------
        int
            The number of rows.
        """
        raise NotImplementedError("Method `columnCount` of FileTreeModel" + " must be implemented by subclasses.")

    def indexPath(self, path: Union[PurePath, PathLike]) -> QModelIndex:
        """
        Construct a `QModelIndex` for the given path.

        If `combine` is enabled, the closest indexed parent path is returned.

        Parameters
        ----------
        path : PurePath or PathLike
            The path to the item the index should point to.

        Returns
        -------
        QModelIndex
            The requested index.
        """
        if isinstance(path, PurePath):
            path = path.parts

        if not path:
            return QModelIndex()  # empty path won't ever be in the model

        # flat mode
        if self.mode == self.DisplayMode.FLAT:
            i = bisect.bisect_left(self._flattened, path)
            if i < len(self._flattened) and self._flattened[i].path == path:
                return self.index(i, 0)
            return QModelIndex()

        # tree mode
        simplified = self.mode == self.DisplayMode.SIMPLIFIED_TREE

        def step(tup, subpath):
            index, i, item = tup

            if not item:
                return index, None

            r, child = item.get(subpath)

            if not child:
                return QModelIndex(), None

            if i <= -1:
                i = r

            if simplified and len(child.children) == 1 and self._simplify_filter(child):
                return index, i, child

            index = self.index(i if simplified else r, 0, index)

            return index, -1, child

        index, i, item = reduce(step, path, (QModelIndex(), -1, self.root))

        return index

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        """
        Construct a `QModelIndex`.

        Returns the index of the item in the model specified by
        the given row, column and parent index.

        Parameters
        ----------
        row : int
        column : int
        parent : QModelIndex, optional

        Returns
        -------
        QModelIndex
            The requested index.
        """
        # different behavior in flat and treemode
        if self.mode == self.DisplayMode.FLAT:
            if 0 <= row < len(self._flattened) and 0 <= column < self.columnCount(parent):
                return self.createIndex(row, column, self._flattened[row])

            return QModelIndex()

        # valid index?
        if not parent.isValid():
            parent_item: FileSystemItem[T] = self.root
        else:
            parent_item = parent.internalPointer()

        item = list(parent_item.children)[row]

        if self.mode == self.DisplayMode.SIMPLIFIED_TREE:
            # combine items with a single child with that child
            while len(item.children) == 1 and self._simplify_filter(item):
                item = item.children[0]

        if 0 <= row < len(parent_item.children) and 0 <= column < self.columnCount(parent):
            return self.createIndex(row, column, item)

        return QModelIndex()

    @overload
    def parent(self, child: QModelIndex) -> QModelIndex:
        pass

    @overload
    def parent(self) -> QObject:
        pass

    def parent(self, child=None):
        """
        Returns the parent of the model item with the given index.

        If the item has no parent, an invalid QModelIndex is returned.
        A common convention used in models that expose tree data structures
        is that only items in the first column have children.
        For that case, when reimplementing this function in a subclass
        the column of the returned QModelIndex would be 0.

        Parameters
        ----------
        child : QModelIndex
            The index of the child item.

        Returns
        -------
        QModelIndex
            The index of the parent item.
        """
        # overloaded variant to retrieve parent of model
        if child is None:
            return super().parent()

        # variant to retrieve parent for data item
        if not child.isValid():
            return QModelIndex()

        # different behavior in tree and flat mode
        if self.mode == self.DisplayMode.FLAT:
            return QModelIndex()  # in flat mode their are no parents

        child_item: FileSystemItem[T] = child.internalPointer()
        parent_item = child_item._parent

        if self.mode == self.DisplayMode.SIMPLIFIED_TREE:
            # combine items with a single child with the child
            while (
                parent_item is not self.root  # do not call filter with root
                and len(parent_item.children) == 1
                and self._simplify_filter(parent_item)
            ):
                parent_item = parent_item._parent

        if parent_item is self.root:
            # Never return root item since it shouldn't be displayed
            return QModelIndex()

        row, item = parent_item._parent.get(parent_item.subpath)
        return self.createIndex(row, 0, parent_item)

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        """
        Returns the item flags for the given index.

        The base class implementation returns a combination of flags
        that enables the item (ItemIsEnabled) and
        allows it to be selected (ItemIsSelectable).

        Parameters
        ----------
        index : QModelIndex
            The index.

        Returns
        -------
        Qt.ItemFlags
            The flags.
        """
        return super().flags(index)

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        """
        Get the data for the given role and section in the given header.

        The header is identified by its orientation.
        For horizontal headers, the section number corresponds to
        the column number. Similarly, for vertical headers,
        the section number corresponds to the row number.

        Parameters
        ----------
        section : int
            The row or column number.
        orientation : Qt.Orientation
            The orientation of the header.
        role : int, optional
            The data role, by default Qt.ItemDataRole.DisplayRole

        Returns
        -------
        Any
            The data for the specified header section.
        """
        return super().headerData(section, orientation, role)


class FileTreeSortProxyModel(QSortFilterProxyModel):
    """
    Sort a FileTreeModel.
    """

    sorted = pyqtSignal(int, Qt.SortOrder)

    def __init__(self, parent=None) -> None:
        """Init."""
        super().__init__(parent)
        self.folders_on_top = False

    @overload
    def keepFoldersOnTop(self) -> bool: ...

    @overload
    def keepFoldersOnTop(self, value: bool) -> bool: ...

    def keepFoldersOnTop(self, value=None) -> bool:
        """
        Set or get whether folders are kept on top when sorting.

        Parameters
        ----------
        value : bool, optional
            The new value, by default None

        Returns
        -------
        bool
            The value of the attribute.
        """
        if value is not None and value != self.folders_on_top:
            self.folders_on_top = value
            # resort
            self.setDynamicSortFilter(False)
            self.sort(self.sortColumn(), self.sortOrder())
            self.setDynamicSortFilter(True)

        return self.folders_on_top

    def extract_path(self, index: QModelIndex):
        """Get the path to compare for a given index."""
        item: FileSystemItem = index.internalPointer()
        model: FileTreeModel = self.sourceModel()

        # name
        if model.mode == FileTreeModel.DisplayMode.FLAT:
            return path_to_str(item.path)

        if model.mode == FileTreeModel.DisplayMode.SIMPLIFIED_TREE:
            parent = index.parent()
            if parent == QModelIndex():
                path = relative_path(model.root.path, item.path)
            else:
                path = relative_path(parent.internalPointer().path, item.path)

            return path[0] if path else ''

        # standard tree mode
        return item.subpath

    def choose_data(self, index: QModelIndex):
        """Choose the data of index used for comparison."""
        raise NotImplementedError(
            "Method `choose_data` of " + "FileTreeSortProxyModel" + " must be implemented by subclasses."
        )

    def lessThan(self, left: QModelIndex, right: QModelIndex) -> bool:
        """
        Return whether the item of `left` is lower than the one of `right`.
        Parameters
        ----------
        left : QModelIndex
            The index left of the `<`.
        right : QModelIndex
            The index right of the `<`.
        Returns
        -------
        bool
            Whether left is lower than right.
        """
        if self.folders_on_top:
            item1 = left.internalPointer()
            item2 = right.internalPointer()
            if item1 is None or item2 is None:
                return False
            ch1 = bool(len(item1.children))
            ch2 = bool(len(item2.children))

            if ch1 ^ ch2:
                if self.sortOrder() == Qt.SortOrder.AscendingOrder:
                    return ch1
                return ch2

        data1 = self.choose_data(left)
        data2 = self.choose_data(right)
        return data1 < data2



from __future__ import annotations

import sys
from datetime import datetime
from typing import NamedTuple

from PyQt6.QtCore import QObject, pyqtSignal


class NetworkStatusMonitor(QObject):
    @classmethod
    def get_network_status_monitor(cls) -> NetworkStatusMonitor:
        if sys.platform == 'darwin':
            from .darwin import DarwinNetworkStatus

            return DarwinNetworkStatus()
        else:
            from .network_manager import (
                DBusException,
                NetworkManagerMonitor,
                UnsupportedException,
            )

            try:
                return NetworkManagerMonitor()
            except (UnsupportedException, DBusException):
                return NullNetworkStatusMonitor()

    network_status_changed = pyqtSignal(bool, name="networkStatusChanged")

    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent)

    def is_network_status_available(self) -> bool:
        """Is the network status really available, and not just a dummy implementation?"""
        return type(self) is not NetworkStatusMonitor

    def is_network_active(self) -> bool:
        """Is there an active network connection.

        True signals that the network is up. The internet may still not be reachable though.
        """
        raise NotImplementedError()

    def is_network_metered(self) -> bool:
        """Is the currently connected network a metered connection?"""
        raise NotImplementedError()

    def get_current_wifi(self) -> str | None:
        """Get current SSID or None if Wifi is off."""
        raise NotImplementedError()

    def get_known_wifis(self) -> list[SystemWifiInfo]:
        """Get WiFi networks known to system."""
        raise NotImplementedError()


class SystemWifiInfo(NamedTuple):
    ssid: str
    last_connected: datetime | None


class NullNetworkStatusMonitor(NetworkStatusMonitor):
    """Dummy implementation, in case we don't have one for current platform."""

    def __init__(self) -> None:
        super().__init__()

    def is_network_active(self) -> bool:
        return True

    def is_network_status_available(self) -> bool:
        return False

    def is_network_metered(self) -> bool:
        return False

    def get_current_wifi(self) -> str | None:
        pass

    def get_known_wifis(self) -> list[SystemWifiInfo]:
        return []



import logging
import re
from html import escape as html_escape

logger = logging.getLogger(__name__)


def escape(text: str) -> str:
    return html_escape(text, quote=False)


def format_richtext(template: str, *args: str) -> str:
    result = template
    for index in range(len(args), 0, -1):
        result = result.replace(f"%{index}", args[index - 1])
    remaining = re.findall(r'%\d+', result)
    if remaining:
        logger.debug(
            'Unreplaced placeholders in richtext template: %s (template=%r, args=%r)', remaining, template, args
        )
    return result


def link(url: str, text: str, color: str = "#0984e3") -> str:
    safe_url = html_escape(url, quote=True)
    safe_text = escape(text)
    return (
        f'<a href="{safe_url}">' f'<span style=" text-decoration: underline; color:{color};">{safe_text}</span>' f"</a>"
    )


def bold(text: str) -> str:
    return f"<b>{escape(text)}</b>"


def italic(text: str) -> str:
    return f'<span style=" font-style:italic;">{escape(text)}</span>'


def code(text: str) -> str:
    return f"<span style=\" font-family:'Courier';\">{escape(text)}</span>"



from __future__ import annotations

from datetime import datetime

import peewee as pw
from playhouse.migrate import Operation, SqliteMigrator, migrate

from .models import (
    DB,
    ArchiveModel,
    BackupProfileModel,
    EventLogModel,
    RepoModel,
    SchemaVersion,
    SettingsModel,
    SourceFileModel,
    WifiSettingModel,
)


def run_migrations(current_schema: SchemaVersion, db_connection: pw.SqliteDatabase) -> None:
    """
    Apply new schema versions to database.

    See http://docs.peewee-orm.com/en/latest/peewee/playhouse.html#schema-migrations
    """
    migrator = SqliteMigrator(db_connection)

    if current_schema.version < 4:  # version 3 to 4
        _apply_schema_update(
            current_schema,
            4,
            migrator.add_column(ArchiveModel._meta.table_name, 'duration', pw.FloatField(null=True)),
            migrator.add_column(ArchiveModel._meta.table_name, 'size', pw.IntegerField(null=True)),
        )
    if current_schema.version < 5:
        _apply_schema_update(
            current_schema,
            5,
            migrator.drop_not_null(WifiSettingModel._meta.table_name, 'last_connected'),
        )

    if current_schema.version < 6:
        _apply_schema_update(
            current_schema,
            6,
            migrator.add_column(EventLogModel._meta.table_name, 'repo_url', pw.CharField(null=True)),
        )

    if current_schema.version < 7:
        _apply_schema_update(
            current_schema,
            7,
            migrator.rename_column(SourceFileModel._meta.table_name, 'config_id', 'profile_id'),
            migrator.drop_column(EventLogModel._meta.table_name, 'profile_id'),
            migrator.add_column(EventLogModel._meta.table_name, 'profile', pw.CharField(null=True)),
        )

    if current_schema.version < 8:
        _apply_schema_update(
            current_schema,
            8,
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'prune_keep_within',
                pw.CharField(null=True),
            ),
        )

    if current_schema.version < 9:
        _apply_schema_update(
            current_schema,
            9,
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'new_archive_name',
                pw.CharField(default="{hostname}-{profile_slug}-{now:%Y-%m-%dT%H:%M:%S}"),
            ),
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'prune_prefix',
                pw.CharField(default="{hostname}-{profile_slug}-"),
            ),
        )

    if current_schema.version < 10:
        _apply_schema_update(
            current_schema,
            10,
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'pre_backup_cmd',
                pw.CharField(default=''),
            ),
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'post_backup_cmd',
                pw.CharField(default=''),
            ),
        )

    if current_schema.version < 11:
        _apply_schema_update(current_schema, 11)
        for profile in BackupProfileModel:
            if profile.compression == 'zstd':
                profile.compression = 'zstd,3'
            if profile.compression == 'lzma,6':
                profile.compression = 'auto,lzma,6'
            profile.save()

    if current_schema.version < 12:
        _apply_schema_update(
            current_schema,
            12,
            migrator.add_column(
                RepoModel._meta.table_name,
                'extra_borg_arguments',
                pw.CharField(default=''),
            ),
        )

    if current_schema.version < 13:
        # Migrate ArchiveModel data to new table to remove unique constraint from snapshot_id column.
        tables = DB.get_tables()
        if ArchiveModel.select().count() == 0 and 'snapshotmodel' in tables:
            cursor = DB.execute_sql('select * from snapshotmodel;')
            fields = [
                ArchiveModel.id,
                ArchiveModel.snapshot_id,
                ArchiveModel.name,
                ArchiveModel.repo,
                ArchiveModel.time,
                ArchiveModel.duration,
                ArchiveModel.size,
            ]
            data = [row for row in cursor.fetchall()]
            with DB.atomic():
                size = 1000
                for i in range(0, len(data), size):
                    ArchiveModel.insert_many(data[i : i + size], fields=fields).execute()

        _apply_schema_update(current_schema, 13)

    if current_schema.version < 14:
        _apply_schema_update(
            current_schema,
            14,
            migrator.add_column(SettingsModel._meta.table_name, 'str_value', pw.CharField(default='')),
        )

    if current_schema.version < 15:
        _apply_schema_update(
            current_schema,
            15,
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'dont_run_on_metered_networks',
                pw.BooleanField(default=True),
            ),
        )

    if current_schema.version < 16:
        _apply_schema_update(
            current_schema,
            16,
            migrator.add_column(
                SourceFileModel._meta.table_name,
                'dir_size',
                pw.BigIntegerField(default=-1),
            ),
            migrator.add_column(
                SourceFileModel._meta.table_name,
                'dir_files_count',
                pw.BigIntegerField(default=-1),
            ),
            migrator.add_column(
                SourceFileModel._meta.table_name,
                'path_isdir',
                pw.BooleanField(default=False),
            ),
        )

    if current_schema.version < 17:
        _apply_schema_update(
            current_schema,
            17,
            migrator.add_column(
                RepoModel._meta.table_name,
                'create_backup_cmd',
                pw.CharField(default=''),
            ),
        )

    if current_schema.version < 18:
        _apply_schema_update(
            current_schema,
            18,
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'schedule_interval_unit',
                pw.CharField(default='hours'),
            ),
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'schedule_interval_count',
                pw.IntegerField(default=3),
            ),
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'schedule_make_up_missed',
                pw.BooleanField(default=False),
            ),
            migrator.add_column(
                EventLogModel._meta.table_name,
                'end_time',
                pw.DateTimeField(default=datetime.now),
            ),
        )

    if current_schema.version < 19:
        _apply_schema_update(
            current_schema,
            19,
            migrator.add_column(SettingsModel._meta.table_name, 'group', pw.CharField(default='')),
        )

    if current_schema.version < 20:
        _apply_schema_update(
            current_schema,
            20,
            migrator.add_column(SettingsModel._meta.table_name, 'tooltip', pw.CharField(default='')),
        )

    if current_schema.version < 21:
        _apply_schema_update(
            current_schema,
            21,
            migrator.add_column(
                ArchiveModel._meta.table_name,
                'trigger',
                pw.CharField(null=True),
            ),
        )

    if current_schema.version < 22:
        _apply_schema_update(
            current_schema,
            22,
            migrator.add_column(
                RepoModel._meta.table_name,
                'name',
                pw.CharField(default=''),
            ),
        )

    if current_schema.version < 23:
        _apply_schema_update(
            current_schema,
            23,
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'compaction_on',
                pw.BooleanField(default=True),
            ),
            migrator.add_column(
                BackupProfileModel._meta.table_name,
                'compaction_weeks',
                pw.IntegerField(default=3),
            ),
        )


def _apply_schema_update(current_schema: SchemaVersion, version_after: int, *operations: Operation) -> None:
    with DB.atomic():
        migrate(*operations)
        current_schema.version = version_after
        current_schema.changed_at = datetime.now()
        current_schema.save()



import logging
import os

from PyQt6 import QtDBus
from PyQt6.QtCore import QMetaType, QVariant
from PyQt6.QtDBus import QDBusMessage
from PyQt6.QtWidgets import QInputDialog

from vorta.i18n import translate
from vorta.keyring.abc import VortaKeyring

logger = logging.getLogger(__name__)


class VortaKWallet5Keyring(VortaKeyring):
    """A wrapper for the qtdbus package to support the custom keyring backend."""

    folder_name = "Vorta"
    service_name = "org.kde.kwalletd5"
    object_path = "/modules/kwalletd5"
    interface_name = "org.kde.KWallet"

    def __init__(self):
        """Initialize the KWallet keyring and test DBus and KDEWallet availability.

        Raises:
            KWalletNotAvailableException: If KWallet is not available or enabled.
        """
        self.iface = QtDBus.QDBusInterface(
            self.service_name,
            self.object_path,
            self.interface_name,
            QtDBus.QDBusConnection.sessionBus(),
        )
        self.handle = -1
        if not (self.iface.isValid() and self.get_result("isEnabled") is True):
            raise KWalletNotAvailableException

    def set_password(self, service, repo_url, password):
        """Set a password in the KWallet.

        Args:
            service: The service name.
            repo_url: The repository URL.
            password: The password to store.
        """
        self.get_result(
            "writePassword",
            args=[self.handle, self.folder_name, repo_url, password, service],
        )
        logger.debug(f"Saved password for repo {repo_url}")

    def delete_password(self, service, repo_url):
        try:
            self.get_result(
                "removeEntry",
                args=[self.handle, self.folder_name, repo_url, service],
            )
            logger.debug(f"Deleted password for repo {repo_url}")
        except Exception as e:
            logger.debug(f"No password to delete for repo {repo_url}: {e}")

    def get_password(self, service, repo_url):
        """Retrieve a password from the KWallet.

        Args:
            service: The service name.
            repo_url: The repository URL.

        Returns:
            str or None: The retrieved password, or None if not found.
        """
        if not (
            self.is_unlocked and self.get_result("hasEntry", args=[self.handle, self.folder_name, repo_url, service])
        ):
            return None
        password = self.get_result("readPassword", args=[self.handle, self.folder_name, repo_url, service])
        logger.debug(translate("KWallet", f"Retrieved password for repo {repo_url}"))
        return password

    def get_result(self, method, args=[]):
        """Call a DBus method and return the raw result.

        Args:
            method: The DBus method to call.
            args: The arguments to pass to the method.

        Returns:
            The raw result from the DBus call, or None on error.
        """
        if args:
            result = self.iface.callWithArgumentList(QtDBus.QDBus.CallMode.AutoDetect, method, args)
        else:
            result = self.iface.call(QtDBus.QDBus.CallMode.AutoDetect, method)

        if result.type() == QDBusMessage.MessageType.ErrorMessage:
            logger.error(translate("KWallet", f"Method '{method}' returned an error message."))
            return None

        arguments = result.arguments()
        return arguments[0] if arguments else None

    @property
    def is_unlocked(self):
        """Check if the wallet is unlocked.

        Returns:
            bool: True if the wallet is unlocked, False otherwise.
        """
        self.try_unlock()
        return self.handle > 0

    def try_unlock(self):
        """Attempt to unlock the wallet.

        Raises:
            ValueError: If the wallet name is invalid or unlocking fails.
        """
        wallet_name = self.get_result("networkWallet")
        if wallet_name is None:
            wallet_name, ok = QInputDialog.getText(
                None,
                translate("KWallet", "Create Wallet"),
                translate("KWallet", "Enter a name for the new wallet:"),
            )
            if not ok or not wallet_name.strip():
                logger.error(
                    translate(
                        "KWallet",
                        "Could not determine a valid wallet name. Aborting unlock attempt.",
                    )
                )
                self.handle = -2
                return

        wId = QVariant(0)
        wId.convert(QMetaType(QMetaType.Type.LongLong.value))
        output = self.get_result("open", args=[wallet_name, wId, "vorta-repo"])
        if output is None:
            logger.error(translate("KWallet", "Failed to open wallet. Aborting unlock attempt."))
            self.handle = -2
            return

        try:
            self.handle = int(output)
        except (ValueError, TypeError):  # For when kwallet is disabled or dbus otherwise broken
            self.handle = -2

    @classmethod
    def get_priority(cls):
        """Get the priority of the keyring.

        Returns:
            int: The priority value.
        """
        return 6 if "KDE" in os.getenv("XDG_CURRENT_DESKTOP", "") else 4

    @property
    def is_system(self):
        """Check if the keyring is a system keyring.

        Returns:
            bool: True if it is a system keyring, False otherwise.
        """
        return True


class KWalletNotAvailableException(Exception):
    """Exception raised when KWallet is not available."""

    pass



import argparse
import errno
import fnmatch
import getpass
import math
import os
import re
import socket
import sys
import unicodedata
import uuid
from datetime import datetime as dt
from datetime import timezone as tz
from functools import reduce
from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar

import psutil
from PyQt6 import QtCore
from PyQt6.QtCore import QFileInfo, QThread, pyqtSignal
from PyQt6.QtWidgets import QApplication, QFileDialog, QSystemTrayIcon

from vorta.borg._compatibility import BorgCompatibility
from vorta.log import logger
from vorta.network_status.abc import NetworkStatusMonitor

# Used to store whether a user wanted to override the
# default directory for the --development flag
DEFAULT_DIR_FLAG = object()
METRIC_UNITS = ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
NONMETRIC_UNITS = ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi']

borg_compat = BorgCompatibility()
_network_status_monitor = None


class FilePathInfoAsync(QThread):
    signal = pyqtSignal(str, str, str)

    def __init__(self, path, exclude_patterns_str):
        self.path = path
        QThread.__init__(self)
        self.exiting = False
        self.exclude_patterns = []
        for _line in (exclude_patterns_str or '').splitlines():
            line = _line.strip()
            if line != '' and not line.startswith("#"):
                self.exclude_patterns.append(line)

    def run(self):
        # logger.info("running thread to get path=%s...", self.path)
        self.size, self.files_count = get_path_datasize(self.path, self.exclude_patterns)
        self.signal.emit(self.path, str(self.size), str(self.files_count))


def normalize_path(path):
    """normalize paths for MacOS (but do nothing on other platforms)"""
    # HFS+ converts paths to a canonical form, so users shouldn't be required to enter an exact match.
    # Windows and Unix filesystems allow different forms, so users always have to enter an exact match.
    return unicodedata.normalize('NFD', path) if sys.platform == 'darwin' else path


# prepare patterns as borg does
# see `FnmatchPattern._prepare` at
# https://github.com/borgbackup/borg/blob/master//src/borg/patterns.py
def prepare_pattern(pattern):
    """Prepare and process fnmatch patterns as borg does"""
    if pattern.endswith(os.path.sep):
        # trailing sep indicates that the contents should be excluded
        # but not the directory it self.
        pattern = os.path.normpath(pattern).rstrip(os.path.sep)
        pattern += os.path.sep + '*' + os.path.sep
    else:
        pattern = os.path.normpath(pattern) + os.path.sep + '*'

    pattern = pattern.lstrip(os.path.sep)  # sep at beginning is removed
    return re.compile(fnmatch.translate(pattern))


def match(pattern: re.Pattern, path: str):
    """Check whether a path matches the given pattern."""
    path = path.lstrip(os.path.sep) + os.path.sep
    return pattern.match(path) is not None


def get_directory_size(dir_path, exclude_patterns):
    '''Get number of files only and total size in bytes from a path.
    Based off https://stackoverflow.com/a/17936789'''
    exclude_patterns = [prepare_pattern(p) for p in exclude_patterns]

    data_size_filtered = 0
    seen = set()
    seen_filtered = set()

    for dir_path, subdirectories, file_names in os.walk(dir_path, topdown=True):
        is_excluded = False
        for pattern in exclude_patterns:
            if match(pattern, dir_path):
                is_excluded = True
                break

        if is_excluded:
            subdirectories.clear()  # so that os.walk won't walk them
            continue

        for file_name in file_names:
            file_path = os.path.join(dir_path, file_name)

            # Ignore symbolic links, since borg doesn't follow them
            if os.path.islink(file_path):
                continue

            is_excluded = False
            for pattern in exclude_patterns:
                if match(pattern, file_path):
                    is_excluded = True
                    break

            try:
                stat = os.stat(file_path)
                if stat.st_ino not in seen:  # Visit each file only once
                    # this won't add the size of a hardlinked file
                    seen.add(stat.st_ino)
                    if not is_excluded:
                        data_size_filtered += stat.st_size
                        seen_filtered.add(stat.st_ino)
            except (FileNotFoundError, PermissionError, OSError):
                # OSError is thrown when the file path name exceeds the maximum length allowed by the operating system
                continue

    files_count_filtered = len(seen_filtered)

    return data_size_filtered, files_count_filtered


def get_network_status_monitor():
    global _network_status_monitor
    if _network_status_monitor is None:
        _network_status_monitor = NetworkStatusMonitor.get_network_status_monitor()
        logger.info(
            'Using %s NetworkStatusMonitor implementation.',
            _network_status_monitor.__class__.__name__,
        )
    return _network_status_monitor


def get_path_datasize(path, exclude_patterns):
    file_info = QFileInfo(path)

    if file_info.isDir():
        data_size, files_count = get_directory_size(file_info.absoluteFilePath(), exclude_patterns)
    else:
        data_size = file_info.size()
        files_count = 1

    return data_size, files_count


def nested_dict():
    """
    Combination of two idioms to quickly build dicts from lists of keys:

    - https://stackoverflow.com/a/16724937/3983708
    - https://stackoverflow.com/a/14692747/3983708
    """
    return dict()


def get_dict_from_list(dataDict, mapList):
    return reduce(lambda d, k: d.setdefault(k, {}), mapList, dataDict)


def choose_file_dialog(parent, title, want_folder=True):
    dialog = QFileDialog(parent, title, os.path.expanduser('~'))
    dialog.setFileMode(QFileDialog.FileMode.Directory if want_folder else QFileDialog.FileMode.ExistingFiles)
    dialog.setParent(parent, QtCore.Qt.WindowType.Sheet)
    if want_folder:
        dialog.setOption(QFileDialog.Option.ShowDirsOnly)
    # Avoid unsupported URL schemes (e.g. smb://) freezing the platform dialog (#1631).
    dialog.setSupportedSchemes(['file'])
    return dialog


def is_ssh_private_key_file(filepath: str) -> bool:
    """Check if the file is a SSH key."""
    try:
        with open(filepath, 'r') as f:
            first_line = f.readline()
        pattern = r'^-----BEGIN(\s\w+)? PRIVATE KEY-----'
        return re.match(pattern, first_line) is not None
    except UnicodeDecodeError:
        return False


def get_private_keys() -> List[str]:
    """Find SSH keys in standard folder."""

    ssh_folder = os.path.expanduser('~/.ssh')

    available_private_keys = []
    if os.path.isdir(ssh_folder):
        for key in os.listdir(ssh_folder):
            key_file = os.path.join(ssh_folder, key)
            if not os.path.isfile(key_file):
                continue
            # ignore config, known_hosts*, *.pub, etc.
            if key.endswith('.pub') or key.startswith('known_hosts') or key == 'config':
                continue
            try:
                if is_ssh_private_key_file(key_file):
                    if os.stat(key_file).st_mode & 0o077 == 0:
                        available_private_keys.append(key)
                    else:
                        logger.warning(f'Permissions for {key_file} are too open.')
                else:
                    logger.debug(f'Not a private SSH key file: {key}')
            except PermissionError:
                logger.warning(f'Permission error while opening file: {key_file}', exc_info=True)
                continue
            except OSError as e:
                if e.errno == errno.ENXIO:
                    # when key_file is a (ControlPath) socket
                    continue
                raise

    return available_private_keys


def sort_sizes(size_list):
    """Sorts sizes with extensions. Assumes that size is already in largest unit possible"""
    final_list = []
    for suffix in [" B", " KB", " MB", " GB", " TB", " PB", " EB", " ZB", " YB"]:
        sub_list = [
            float(size[: -len(suffix)])
            for size in size_list
            if size.endswith(suffix) and size[: -len(suffix)][-1].isnumeric()
        ]
        sub_list.sort()
        final_list += [(str(size) + suffix) for size in sub_list]
        # Skip additional loops
        if len(final_list) == len(size_list):
            break
    return final_list


Number = TypeVar("Number", int, float)


def clamp(n: Number, min_: Number, max_: Number) -> Number:
    """Restrict the number n inside a range"""
    return min(max_, max(n, min_))


def find_best_unit_for_sizes(sizes: Iterable[int], metric: bool = True, precision: int = 1) -> int:
    """
    Selects the index of the biggest unit (see the lists in the pretty_bytes function) capable of
    representing the smallest size in the sizes iterable.
    """
    min_size = min((s for s in sizes if isinstance(s, int)), default=None)
    return find_best_unit_for_size(min_size, metric=metric, precision=precision)


def find_best_unit_for_size(size: Optional[int], metric: bool = True, precision: int = 1) -> int:
    """
    Selects the index of the biggest unit (see the lists in the pretty_bytes function) capable of
    representing the passed size.
    """
    if not isinstance(size, int) or size == 0:  # this will also take care of the None case
        return 0
    power = 10**3 if metric else 2**10
    n = math.floor(math.log(abs(size) * 10**precision, power))
    return n


def pretty_bytes(
    size: int, metric: bool = True, sign: bool = False, precision: int = 1, fixed_unit: Optional[int] = None
) -> str:
    """
    Formats the size with the requested unit and precision. The find_best_size_unit function
    can be used to find the correct unit for a list of sizes. If no fixed_unit is passed it will
    find the biggest unit to represent the size
    """
    if not isinstance(size, int):
        return ''
    prefix = '+' if sign and size > 0 else ''
    power, units = (10**3, METRIC_UNITS) if metric else (2**10, NONMETRIC_UNITS)
    if fixed_unit is None:
        n = find_best_unit_for_size(size, metric=metric, precision=precision)
    else:
        n = fixed_unit
    n = clamp(n, 0, len(units) - 1)
    size /= power**n
    try:
        unit = units[n]
        digits = f'%.{precision}f' % (round(size, precision))
        return f'{prefix}{digits} {unit}B'
    except KeyError as error:
        logger.error(error)
        return "NaN"


def get_asset(path):
    if getattr(sys, 'frozen', False):
        # we are running in a bundle
        bundle_dir = os.path.join(sys._MEIPASS, 'assets')
    else:
        # we are running in a normal Python environment
        bundle_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'assets')
    return os.path.join(bundle_dir, path)


def get_sorted_wifis(profile):
    """
    Get Wifi networks known to the OS (only current one on macOS) and
    merge with networks from other profiles. Update last connected time.
    """

    from vorta.store.models import WifiSettingModel

    # Pull networks known to OS and all other backup profiles
    system_wifis = get_network_status_monitor().get_known_wifis()
    from_other_profiles = WifiSettingModel.select().where(WifiSettingModel.profile != profile.id).execute()

    for wifi in list(from_other_profiles) + system_wifis:
        db_wifi, created = WifiSettingModel.get_or_create(
            ssid=wifi.ssid,
            profile=profile.id,
            defaults={'last_connected': wifi.last_connected, 'allowed': True},
        )

        # Update last connected time
        if not created and db_wifi.last_connected != wifi.last_connected:
            db_wifi.last_connected = wifi.last_connected
            db_wifi.save()

    # Finally return list of networks and settings for that profile
    return (
        WifiSettingModel.select()
        .where(WifiSettingModel.profile == profile.id)
        .order_by(-WifiSettingModel.last_connected)
    )


def parse_args():
    parser = argparse.ArgumentParser(description='Vorta Backup GUI for Borg.')
    parser.add_argument('--version', '-V', action='store_true', help="Show version and exit.")
    parser.add_argument(
        '--daemonize',
        '-d',
        action='store_true',
        help="Fork to background and don't open window on startup.",
    )
    parser.add_argument(
        '--create',
        dest='profile',
        help='Create a backup in the background using the given profile. '
        'Vorta must already be running for this to work.',
    )
    # the "development" attribute will be None if the flag is not called
    # if the flag is called without an extra argument, the "development" attribute
    # will be set to the value of DEFAULT_DIR_FLAG.
    # if the flag is called with an extra argument, the "development" attribute
    # will be set to that argument
    parser.add_argument(
        '--development',
        '-D',
        nargs='?',
        const=DEFAULT_DIR_FLAG,
        metavar="CONFIG_DIRECTORY",
        help='Start vorta in a local development environment. '
        'All log, config, cache, and temp files will be stored within the project tree. '
        'You can follow this flag with an optional path and it will store the files in the provided location.',
    )
    return parser.parse_known_args()[0]


def slugify(value):
    """
    Converts to lowercase, removes non-word characters (alphanumerics and
    underscores) and converts spaces to hyphens. Also strips leading and
    trailing whitespace.

    Copied from Django.
    """
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    return re.sub(r'[-\s]+', '-', value)


def uses_dark_mode():
    """
    This function detects if we are running in dark mode (e.g. macOS dark mode).
    """
    palette = QApplication.instance().palette()
    return palette.windowText().color().lightness() > palette.window().color().lightness()


# patched socket.getfqdn() - see https://bugs.python.org/issue5004
# Reused with permission from https://github.com/borgbackup/borg/blob/master/src/borg/platform/base.py (BSD-3-Clause)
def _getfqdn(name=""):
    """Get fully qualified domain name from name.
    An empty argument is interpreted as meaning the local host.
    """
    name = name.strip()
    if not name or name == "0.0.0.0":
        name = socket.gethostname()
    try:
        addrs = socket.getaddrinfo(name, None, 0, socket.SOCK_DGRAM, 0, socket.AI_CANONNAME)
    except OSError:
        pass
    else:
        for addr in addrs:
            if addr[3]:
                name = addr[3]
                break
    return name


def format_archive_name(profile, archive_name_tpl):
    """
    Generate an archive name. Default set in models.BackupProfileModel
    """
    hostname = socket.gethostname()
    hostname = hostname.split(".")[0]
    borg_version = borg_compat.get_version()[0]  # Getting only borg version
    borg_version_tuple = tuple(borg_version.split("."))
    available_vars = {
        "pid": os.getpid(),
        'hostname': hostname,
        'fqdn': _getfqdn(hostname),
        "reverse-fqdn": ".".join(reversed(_getfqdn(hostname).split("."))),
        'profile_id': profile.id,
        'profile_slug': profile.slug(),
        'now': dt.now(),
        'utc_now': dt.now(tz.utc),
        'utcnow': dt.now(tz.utc),
        'user': getpass.getuser(),
        "uuid4": str(uuid.uuid4()),
        'borgversion': borg_version,
        'borgmajor': "%s" % borg_version_tuple[:1],
        'borgminor': "%s.%s" % borg_version_tuple[:2],
        'borgpatch': "%s.%s.%s" % borg_version_tuple[:3],
    }
    return archive_name_tpl.format(**available_vars)


SHELL_PATTERN_ELEMENT = re.compile(r'([?\[\]*])')


def is_mount_command(proc):
    try:
        name = proc.name()
        if name == 'borg' or name.startswith('python'):
            if 'mount' in proc.cmdline():
                return True
    except (psutil.ZombieProcess, psutil.AccessDenied, psutil.NoSuchProcess):
        pass
    return False


def extract_mount_points_v2(proc, repo_url):
    mount_points = {}
    repo_mounts = []

    cmd = proc.cmdline()
    if repo_url in cmd:
        i = cmd.index(repo_url)
        if len(cmd) > i + 1:
            mount_point = cmd[i + 1]

            ao = '-a' in cmd
            if ao or '--match-archives' in cmd:
                i = cmd.index('-a' if ao else '--match-archives')
                if len(cmd) >= i + 1 and not SHELL_PATTERN_ELEMENT.search(cmd[i + 1]):
                    mount_points[mount_point] = cmd[i + 1]
            else:
                repo_mounts.append(mount_point)

    return mount_points, repo_mounts


def extract_mount_points_v1(proc, repo_url):
    mount_points = {}
    repo_mounts = []

    for idx, parameter in enumerate(proc.cmdline()):
        if parameter.startswith(repo_url):
            if len(proc.cmdline()) > idx + 1:
                mount_point = proc.cmdline()[idx + 1]

                if parameter[len(repo_url) :].startswith('::'):
                    archive_name = parameter[len(repo_url) + 2 :]
                    mount_points[archive_name] = mount_point
                else:
                    repo_mounts.append(mount_point)

    return mount_points, repo_mounts


def get_mount_points(repo_url):
    mount_points = {}
    repo_mounts = []

    for proc in psutil.process_iter():
        if not is_mount_command(proc):
            continue

        if borg_compat.check('V2'):
            v2_mount_points, v2_repo_mounts = extract_mount_points_v2(proc, repo_url)
            mount_points.update(v2_mount_points)
            repo_mounts.extend(v2_repo_mounts)
        else:
            v1_mount_points, v1_repo_mounts = extract_mount_points_v1(proc, repo_url)
            mount_points.update(v1_mount_points)
            repo_mounts.extend(v1_repo_mounts)

    return mount_points, repo_mounts


def is_system_tray_available():
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
        tray = QSystemTrayIcon()
        is_available = tray.isSystemTrayAvailable()
        app.quit()
    else:
        tray = QSystemTrayIcon()
        is_available = tray.isSystemTrayAvailable()

    return is_available


def search(key, iterable: Iterable, func: Callable = None) -> Tuple[int, Any]:
    """
    Search for a key in an iterable.

    Before comparing an item with the key `func` is called on the item.

    Parameters
    ----------
    key : Any
        The key to search for.
    iterable : Iterable
        The iterable to search in.
    func : Callable, optional
        The function to apply, by default None

    Returns
    -------
    Tuple[int, Any] or None
        The index and the item in case of a match else `None`.
    """
    if not func:

        def func(x):
            return x

    for i, item in enumerate(iterable):
        if func(item) == key:
            return i, item

    return None



from datetime import datetime as dt

from vorta.store.models import ArchiveModel, RepoModel
from vorta.utils import borg_compat

from .borg_job import BorgJob


class BorgListRepoJob(BorgJob):
    def started_event(self):
        self.app.backup_started_event.emit()
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Refreshing archives…')}")

    def finished_event(self, result):
        self.app.backup_finished_event.emit(result)
        self.result.emit(result)
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Refreshing archives done.')}")

    @classmethod
    def prepare(cls, profile):
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret
        else:
            ret['ok'] = False  # Set back to false, so we can do our own checks here.

        if borg_compat.check('V2'):
            cmd = ['borg', 'repo-list', '--info', '--log-json', '--json', '-r']
        else:
            cmd = ['borg', 'list', '--info', '--log-json', '--json']
        cmd.append(f'{profile.repo.url}')

        ret['ok'] = True
        ret['cmd'] = cmd

        return ret

    def process_result(self, result):
        if result['returncode'] == 0:
            repo, created = RepoModel.get_or_create(url=result['cmd'][-1])
            if not result['data']:
                result['data'] = {}  # TODO: Workaround for tests. Can't read mock results 2x.
            remote_archives = result['data'].get('archives', [])

            # Delete archives that don't exist on the remote side
            for archive in ArchiveModel.select().where(ArchiveModel.repo == repo.id):
                if not list(filter(lambda s: s['id'] == archive.snapshot_id, remote_archives)):
                    archive.delete_instance()

            # Add remote archives we don't have locally.
            for archive in result['data'].get('archives', []):
                new_archive, _ = ArchiveModel.get_or_create(
                    snapshot_id=archive['id'],
                    repo=repo.id,
                    defaults={
                        'name': archive['name'],
                        # Convert to local time (for Borg 2.x UTC timestamps) before storing as naive datetime
                        'time': dt.fromisoformat(archive['time']).astimezone().replace(tzinfo=None),
                    },
                )
                new_archive.save()



from __future__ import annotations

import enum
import logging
import threading
from datetime import datetime as dt
from datetime import timedelta
from typing import Any, NamedTuple

from packaging import version
from PyQt6 import QtCore, QtDBus
from PyQt6.QtCore import QTimer
from PyQt6.QtWidgets import QApplication

from vorta import application
from vorta.borg.check import BorgCheckJob
from vorta.borg.compact import BorgCompactJob
from vorta.borg.create import BorgCreateJob
from vorta.borg.list_repo import BorgListRepoJob
from vorta.borg.prune import BorgPruneJob
from vorta.i18n import translate
from vorta.notifications import VortaNotifications
from vorta.store.models import BackupProfileModel, EventLogModel
from vorta.utils import borg_compat, get_network_status_monitor

logger = logging.getLogger(__name__)


class ScheduleStatusType(enum.Enum):
    SCHEDULED = enum.auto()  # date provided
    UNSCHEDULED = enum.auto()  # Unknown
    TOO_FAR_AHEAD = enum.auto()  # QTimer range exceeded, date provided
    NO_PREVIOUS_BACKUP = enum.auto()  # run a manual backup first


class ScheduleStatus(NamedTuple):
    type: ScheduleStatusType
    time: dt | None = None


class VortaScheduler(QtCore.QObject):
    #: The schedule for a profile changed.
    schedule_changed = QtCore.pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        #: mapping of profiles to timers
        self.timers: dict[int, dict[str, QTimer | dt | ScheduleStatusType | None]] = dict()

        self.app: application.VortaApp = QApplication.instance()
        self.lock = threading.Lock()

        # pausing will prevent scheduling for a specified time
        self.pauses: dict[int, tuple[dt, QtCore.QTimer]] = dict()

        # Set additional timer to make sure background tasks stay scheduled.
        # E.g. after hibernation
        self.qt_timer = QTimer()
        self.qt_timer.timeout.connect(self.reload_all_timers)
        self.qt_timer.setInterval(15 * 60 * 1000)
        self.qt_timer.start()

        # connect signals
        self.app.backup_finished_event.connect(lambda res: self.set_timer_for_profile(res['params']['profile_id']))

        # Connect to network manager to monitor net status
        self.net_status = get_network_status_monitor()
        self.net_status.network_status_changed.connect(self.networkStatusChanged)
        self._net_up = self.net_status.is_network_active()

        # connect to `systemd-logind` to receive sleep/resume events
        # The signal `PrepareForSleep` will be emitted before and after hibernation.
        service = "org.freedesktop.login1"
        path = "/org/freedesktop/login1"
        interface = "org.freedesktop.login1.Manager"
        name = "PrepareForSleep"
        bus = QtDBus.QDBusConnection.systemBus()
        if bus.isConnected() and bus.interface().isServiceRegistered(service).value():
            self.bus = bus
            self.bus.connect(service, path, interface, name, "b", self.loginSuspendNotify)
        else:
            logger.warning('Failed to connect to DBUS interface to detect sleep/resume events')

    @QtCore.pyqtSlot(bool)
    def loginSuspendNotify(self, suspend: bool) -> None:
        if not suspend:
            logger.debug("Got login suspend/resume notification")
            # Defensively refetch in case the network status didn't arrive
            self._net_up = self.net_status.is_network_active()
            self.reload_all_timers()

    @QtCore.pyqtSlot(bool)
    def networkStatusChanged(self, up: bool):
        reload = self._net_up != up
        self._net_up = up
        logger.debug(f"network status up={up}")
        if reload:
            logger.info("updating schedule due to network status change")
            self.reload_all_timers()

    def tr(self, *args: Any, **kwargs: Any) -> str:
        scope = self.__class__.__name__
        return translate(scope, *args, **kwargs)

    def pause(self, profile_id: int, until: dt | None = None) -> None:
        """
        Call a timeout for scheduling of a given profile.

        If `until` is omitted, a default time for the break is calculated.

        .. warning::
            This method won't work correctly when called from a non-`QThread`.


        Parameters
        ----------
        profile_id : int
            The profile to pause the scheduling for.
        until : dt | None, optional
            The time to end the pause, by default None
        """
        profile = BackupProfileModel.get_or_none(id=profile_id)
        if profile is None:  # profile doesn't exist any more.
            return

        if profile.schedule_mode == 'off':
            return

        if until is None:
            # calculate default timeout

            if profile.schedule_mode == 'interval':
                interval = timedelta(**{profile.schedule_interval_unit: profile.schedule_interval_count})
            else:
                # fixed
                interval = timedelta(days=1)

            timeout = interval // 6  # 60 / 6 = 10 [min]
            timeout = max(min(timeout, timedelta(hours=1)), timedelta(minutes=1))  # 1 <= t <= 60

            until = dt.now().replace(microsecond=0) + timeout
        elif until < dt.now():
            return

        # remove existing schedule
        self.remove_job(profile_id)

        # setting timer for reschedule is not possible if called
        # from a non-QThread -  it won't fail but won't work
        timer_value = max(1, (until - dt.now()).total_seconds())
        timer = QtCore.QTimer()
        timer.setInterval(int(timer_value * 1000) + 100)
        timer.timeout.connect(lambda: self.set_timer_for_profile(profile_id))
        timer.start()

        # set timeout/pause
        other_pause = self.pauses.get(profile_id)
        if other_pause is not None:
            logger.debug(f"Override existing timeout for profile {profile_id}")

        self.pauses[profile_id] = (until, timer)
        logger.debug(f"Paused {profile_id} until {until.strftime('%Y-%m-%d %H:%M:%S')}")

    def unpause(self, profile_id: int) -> None:
        """
        Return to scheduling for a profile.

        Parameters
        ----------
        profile_id : int
            The profile to end the timeout for.
        """
        profile = BackupProfileModel.get_or_none(id=profile_id)
        if profile is None:  # profile doesn't exist any more.
            return

        pause = self.pauses.get(profile_id)
        if pause is None:  # already unpaused
            return

        dummy, timer = pause
        timer.stop()
        del self.pauses[profile_id]

        logger.debug(f"Unpaused {profile_id}")

        self.set_timer_for_profile(profile_id)

    def paused(self, profile_id: int) -> bool:
        """
        Determine whether scheduling for a profile is paused

        Parameters
        ----------
        profile_id : int

        Returns
        -------
        bool
        """
        return self.pauses.get(profile_id) is not None

    def set_timer_for_profile(self, profile_id: int) -> None:
        """
        Set a timer for next scheduled backup run of this profile.

        Removes existing jobs if set to manual only or no repo is assigned.

        Else will look for previous scheduled backups and catch up if
        schedule_make_up_missed is enabled.

        Or, if catch-up is not enabled, will add interval to last run to find
        next suitable backup time.
        """
        profile = BackupProfileModel.get_or_none(id=profile_id)
        logger.debug('Profile: %s, %d %d', str(profile), profile.schedule_fixed_hour, profile.schedule_fixed_minute)
        if profile is None:  # profile doesn't exist any more.
            return

        with self.lock:  # Acquire lock
            self.remove_job(profile_id)  # reset schedule

            pause = self.pauses.get(profile_id)
            if pause is not None:
                pause_end, timer = pause
                if dt.now() < pause_end:
                    logger.debug(
                        'Nothing scheduled for profile %s ' + 'because of timeout until %s.',
                        profile_id,
                        pause[0].strftime('%Y-%m-%d %H:%M:%S'),
                    )
                    return
                else:
                    timer.stop()
                    del self.pauses[profile_id]

            if profile.repo is None:  # No backups without repo set
                logger.debug(
                    'Nothing scheduled for profile %s because of unset repo.',
                    profile_id,
                )
                # Emit signal so that e.g. the GUI can react to the new schedule
                self.schedule_changed.emit()
                return

            if profile.schedule_mode == 'off':
                logger.debug('Scheduler for profile %s is disabled.', profile_id)
                # Emit signal so that e.g. the GUI can react to the new schedule
                self.schedule_changed.emit()
                return

            logger.info('Setting timer for profile %s', profile_id)

            # determine last backup time
            last_run_log = (
                EventLogModel.select()
                .where(
                    EventLogModel.subcommand == 'create',
                    EventLogModel.category == 'scheduled',
                    EventLogModel.profile == profile.id,
                    0 <= EventLogModel.returncode <= 1,
                )
                .order_by(EventLogModel.end_time.desc())
                .first()
            )

            if last_run_log is None:
                # look for non scheduled (manual) backup runs
                last_run_log = (
                    EventLogModel.select()
                    .where(
                        EventLogModel.subcommand == 'create',
                        EventLogModel.profile == profile.id,
                        0 <= EventLogModel.returncode <= 1,
                    )
                    .order_by(EventLogModel.end_time.desc())
                    .first()
                )

            if last_run_log is None:
                logger.info(
                    f"Nothing scheduled for profile {profile_id} "
                    + "because it would be the first backup "
                    + "for this profile."
                )
                self.timers[profile_id] = {'type': ScheduleStatusType.NO_PREVIOUS_BACKUP}
                # Emit signal so that e.g. the GUI can react to the new schedule
                self.schedule_changed.emit()
                return

            # calculate next scheduled time
            if profile.schedule_mode == 'interval':
                last_time: dt = last_run_log.end_time

                interval = {profile.schedule_interval_unit: profile.schedule_interval_count}
                next_time = last_time + timedelta(**interval)

            elif profile.schedule_mode == 'fixed':
                last_time = last_run_log.end_time

                next_time = last_time.replace(
                    hour=profile.schedule_fixed_hour,
                    minute=profile.schedule_fixed_minute,
                    second=0,
                    microsecond=0,
                ) + timedelta(days=1)

            else:
                # unknown schedule mode
                raise ValueError("Unknown schedule mode '{}'".format(profile.schedule_mode))

            logger.debug('Last run time: %s', last_time)

            needs_network = profile.repo is not None and profile.repo.is_remote_repo()
            # handle missing of a scheduled time
            if next_time <= dt.now():
                if profile.schedule_make_up_missed and (self._net_up or not needs_network):
                    self.lock.release()
                    try:
                        logger.debug(
                            'Catching up by running job for %s (%s)',
                            profile.name,
                            profile_id,
                        )
                        self.create_backup(profile_id)
                    finally:
                        self.lock.acquire()  # with-statement will try to release

                    return  # create_backup will lead to a call to this method
                elif profile.schedule_make_up_missed and not self._net_up and needs_network:
                    logger.debug('Skipping catchup %s (%s), the network is not available', profile.name, profile.id)

                # calculate next time from now
                if profile.schedule_mode == 'interval':
                    # next_time % interval should be 0
                    # while next_time > now
                    delta = dt.now() - last_time
                    next_time = dt.now() - delta % timedelta(**interval)
                    next_time += timedelta(**interval)

                elif profile.schedule_mode == 'fixed':
                    # schedule for today
                    next_time = dt.now().replace(
                        hour=profile.schedule_fixed_hour,
                        minute=profile.schedule_fixed_minute,
                        second=0,
                        microsecond=0,
                    )

                    if next_time <= dt.now():
                        # time for today has passed, schedule for tomorrow
                        next_time += timedelta(days=1)

            # start QTimer
            timer_ms = (next_time - dt.now()).total_seconds() * 1000

            if timer_ms < 2**31 - 1:
                logger.debug('Scheduling next run for %s', next_time)

                timer = QTimer()
                timer.setSingleShot(True)
                timer.setInterval(int(timer_ms))
                timer.timeout.connect(lambda: self.create_backup(profile_id))
                timer.start()

                self.timers[profile_id] = {
                    'qtt': timer,
                    'dt': next_time,
                    'type': ScheduleStatusType.SCHEDULED,
                }
            else:
                # int to big to pass it to qt which expects a c++ int
                # wait 15 min for regular reschedule
                logger.debug(f"Couldn't schedule for {next_time} because " f"timer value {timer_ms} too large.")

                self.timers[profile_id] = {
                    'dt': next_time,
                    'type': ScheduleStatusType.TOO_FAR_AHEAD,
                }

        # Emit signal so that e.g. the GUI can react to the new schedule
        self.schedule_changed.emit()

    def reload_all_timers(self) -> None:
        logger.debug('Refreshing all scheduler timers')
        for profile in BackupProfileModel.select():
            # Only set a timer for the profile if the network is actually up
            if profile.repo is None:
                logger.debug("nothing scheduled for %s because of unset repo", profile.id)
            elif not profile.repo.is_remote_repo() or self._net_up:
                logger.debug("scheduling %s", profile.id)
                self.set_timer_for_profile(profile.id)
            else:
                logger.debug("Network is down, not scheduling %s", profile.id)
                self.remove_job(profile.id)

    def next_job(self) -> str:
        now = dt.now()

        def is_scheduled(timer):
            return timer["type"] == ScheduleStatusType.SCHEDULED and timer["qtt"].isActive() and timer["dt"] >= now

        scheduled = {profile_id: timer for profile_id, timer in self.timers.items() if is_scheduled(timer)}
        if len(scheduled) == 0:
            return self.tr("None scheduled")

        closest_job = min(scheduled.items(), key=lambda item: item[1]["dt"])
        profile_id, timer = closest_job
        time = timer["dt"]
        profile = BackupProfileModel.get_or_none(id=profile_id)

        time_format = "%H:%M"
        if time - now > timedelta(days=1):
            time_format = "%b %d, %H:%M"
        return f"{time.strftime(time_format)} ({profile.name})"

    def next_job_for_profile(self, profile_id: int) -> ScheduleStatus:
        job = self.timers.get(profile_id)
        if job is None:
            return ScheduleStatus(ScheduleStatusType.UNSCHEDULED)
        return ScheduleStatus(job['type'], time=job.get('dt'))  # type: ignore[arg-type]

    def create_backup(self, profile_id: int) -> None:
        notifier = VortaNotifications.pick()
        profile = BackupProfileModel.get_or_none(id=profile_id)

        if profile is None:
            logger.info('Profile not found. Maybe deleted?')
            return

        # Skip if a job for this profile (repo) is already in progress
        if self.app.jobs_manager.is_worker_running(site=profile.repo.id):
            logger.debug('A job for repo %s is already active.', profile.repo.id)
            self.pause(profile_id)
            return

        with self.lock:
            logger.info('Starting background backup for %s', profile.name)
            notifier.deliver(
                self.tr('Vorta Backup'),
                self.tr('Starting background backup for %s.') % profile.name,
                level='info',
            )
            msg = BorgCreateJob.prepare(profile)
            if msg['ok']:
                logger.info('Preparation for backup successful.')
                msg['category'] = 'scheduled'
                job = BorgCreateJob(msg['cmd'], msg, profile.repo.id)
                job.result.connect(self.notify)
                self.app.jobs_manager.add_job(job)
            else:
                # Default to 'error': unexpected failures notify.
                # Expected skips (WiFi/metered) use 'info' to suppress.
                level = msg.get('level', 'error')
                if level == 'error':
                    logger.error('Conditions for backup not met. Aborting.')
                    logger.error(msg['message'])
                    notifier.deliver(
                        self.tr('Vorta Backup'),
                        translate('messages', msg['message']),
                        level='error',
                    )
                else:
                    logger.info('Backup skipped: %s', msg['message'])
                self.pause(profile_id)

    def notify(self, result: dict[str, Any]) -> None:
        notifier = VortaNotifications.pick()
        profile_name = result['params']['profile_name']
        profile_id = result['params']['profile'].id

        if result['returncode'] in [0, 1]:
            notifier.deliver(
                self.tr('Vorta Backup'),
                self.tr('Backup successful for %s.') % profile_name,
                level='info',
            )
            logger.info('Backup creation successful.')
            # unpause scheduler
            self.unpause(result['params']['profile_id'])

            self.post_backup_tasks(profile_id)
        else:
            notifier.deliver(
                self.tr('Vorta Backup'),
                self.tr('Error during backup creation for %s.') % profile_name,
                level='error',
            )
            logger.error('Error during backup creation.')
            # pause scheduler
            # if a scheduled backup fails the scheduler should pause
            # temporarily.
            self.pause(result['params']['profile_id'])

        self.set_timer_for_profile(profile_id)

    def post_backup_tasks(self, profile_id: int) -> None:
        """
        Pruning and checking after successful backup.
        """
        profile = BackupProfileModel.get(id=profile_id)
        notifier = VortaNotifications.pick()
        logger.info('Doing post-backup jobs for %s', profile.name)
        if profile.prune_on:
            msg = BorgPruneJob.prepare(profile)
            if msg['ok']:
                job = BorgPruneJob(msg['cmd'], msg, profile.repo.id)
                self.app.jobs_manager.add_job(job)

                # Refresh archives
                msg = BorgListRepoJob.prepare(profile)
                if msg['ok']:
                    job = BorgListRepoJob(msg['cmd'], msg, profile.repo.id)
                    self.app.jobs_manager.add_job(job)

        validation_cutoff = dt.now() - timedelta(days=7 * profile.validation_weeks)
        recent_validations = (
            EventLogModel.select()
            .where(
                (EventLogModel.subcommand == 'check')
                & (EventLogModel.start_time > validation_cutoff)
                & (EventLogModel.repo_url == profile.repo.url)
            )
            .count()
        )
        if profile.validation_on and recent_validations == 0:
            msg = BorgCheckJob.prepare(profile)
            if msg['ok']:
                job = BorgCheckJob(msg['cmd'], msg, profile.repo.id)
                self.app.jobs_manager.add_job(job)

        compaction_cutoff = dt.now() - timedelta(days=7 * profile.compaction_weeks)
        recent_compactions = (
            EventLogModel.select()
            .where(
                (EventLogModel.subcommand == '--info')
                & (EventLogModel.start_time > compaction_cutoff)
                & (EventLogModel.repo_url == profile.repo.url)
            )
            .count()
        )

        if (
            profile.compaction_on
            and recent_compactions == 0
            and version.parse(borg_compat.version) >= version.parse("1.2")
        ):
            msg = BorgCompactJob.prepare(profile)
            if msg['ok']:
                job = BorgCompactJob(msg['cmd'], msg, profile.repo.id)
                self.app.jobs_manager.add_job(job)

        logger.info('Finished background task for profile %s', profile.name)
        notifier.deliver(
            self.tr('Vorta Backup'),
            self.tr('Post Backup Tasks successful for %s' % profile.name),
            level='info',
        )

    def remove_job(self, profile_id: int) -> None:
        if profile_id in self.timers:
            qtimer = self.timers[profile_id].get('qtt')
            if qtimer is not None:
                qtimer.stop()

            del self.timers[profile_id]



from typing import Any, Dict

from vorta import config
from vorta.i18n import translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.utils import borg_compat

from .borg_job import BorgJob


class BorgCompactJob(BorgJob):
    def started_event(self):
        self.app.backup_started_event.emit()
        self.app.backup_progress_event.emit(
            f"[{self.params['profile_name']} {self.tr('Starting repository compaction...')}]"
        )

    def finished_event(self, result: Dict[str, Any]):
        """
        Process that the job terminated with the given results.

        Parameters
        ----------
        result : Dict[str, Any]
            The (json-like) dictionary containing the job results.
        """
        self.app.backup_finished_event.emit(result)
        self.result.emit(result)
        if result['returncode'] != 0:
            logs_link = link(config.LOG_DIR.as_uri(), translate('messages', 'logs'))
            message = format_richtext(
                escape(translate('BorgCompactJob', 'Errors during compaction. See the %1 for details.')),
                logs_link,
            )
            self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] " + message)
        else:
            self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Compaction completed.')}")

    @classmethod
    def prepare(cls, profile):
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret
        else:
            ret['ok'] = False  # Set back to false, so we can do our own checks here.

        if not borg_compat.check('COMPACT_SUBCOMMAND'):
            raise Exception('The compact action needs Borg >= 1.2.0')

        cmd = ['borg', '--info', '--log-json', 'compact', '--progress']
        if borg_compat.check('V2'):
            cmd = cmd + ["-r", profile.repo.url]
        else:
            cmd.append(f'{profile.repo.url}')

        ret['ok'] = True
        ret['cmd'] = cmd

        return ret



import datetime
import json
from json import JSONDecodeError

from playhouse.shortcuts import dict_to_model, model_to_dict

from vorta.keyring.abc import VortaKeyring
from vorta.store.connection import DB, SCHEMA_VERSION, init_db
from vorta.store.models import (
    BackupProfileModel,
    ExclusionModel,
    RepoModel,
    SchemaVersion,
    SettingsModel,
    SourceFileModel,
    WifiSettingModel,
)


class ProfileExport:
    def __init__(self, profile_dict):
        self._profile_dict = profile_dict

    @property
    def id(self):
        return self._profile_dict['id']

    @property
    def name(self):
        return self._profile_dict['name']

    @property
    def schema_version(self):
        return self._profile_dict['SchemaVersion']['version']

    @property
    def repo_url(self):
        if (
            'repo' in self._profile_dict
            and isinstance(self._profile_dict['repo'], dict)
            and 'url' in self._profile_dict['repo']
        ):
            return self._profile_dict['repo']['url']
        return None

    @property
    def repo_password(self):
        return self._profile_dict['password'] if 'password' in self._profile_dict else None

    @repo_password.setter
    def repo_password(self, password):
        self._profile_dict['password'] = password

    @classmethod
    def from_db(cls, profile, store_password=True, include_settings=True):
        profile_dict = model_to_dict(profile, exclude=[RepoModel.id])  # Have to retain profile ID

        keyring = VortaKeyring.get_keyring()
        if store_password:
            profile_dict['password'] = keyring.get_password('vorta-repo', profile.repo.url)

        # For all below, exclude ids to prevent collisions. DB will automatically reassign ids
        # Add SourceFileModel
        profile_dict['SourceFileModel'] = [
            model_to_dict(source, recurse=False, exclude=[SourceFileModel.id])
            for source in SourceFileModel.select().where(SourceFileModel.profile == profile)
        ]
        # Add ExclusionModel
        profile_dict['ExclusionModel'] = [
            model_to_dict(exclusion, recurse=False, exclude=[ExclusionModel.id])
            for exclusion in ExclusionModel.select().where(ExclusionModel.profile == profile)
        ]
        # Add SchemaVersion
        profile_dict['SchemaVersion'] = model_to_dict(SchemaVersion.get(id=1))

        if include_settings:
            # Add WifiSettingModel
            profile_dict['WifiSettingModel'] = [
                model_to_dict(wifi, recurse=False)
                for wifi in WifiSettingModel.select().where(WifiSettingModel.profile == profile.id)
            ]
            # Add SettingsModel
            profile_dict['SettingsModel'] = [model_to_dict(s, exclude=[SettingsModel.id]) for s in SettingsModel]
        return ProfileExport(profile_dict)

    def to_db(self, overwrite_profile=False, overwrite_settings=True):
        profile_schema = self._profile_dict['SchemaVersion']['version']
        keyring = VortaKeyring.get_keyring()
        if SCHEMA_VERSION < profile_schema:
            raise VersionException()
        elif SCHEMA_VERSION > profile_schema:
            # Add model upgrading code here, only needed if not adding columns
            if profile_schema < 16:
                for sourcedir in self._profile_dict['SourceFileModel']:
                    sourcedir['dir_files_count'] = -1
                    sourcedir['dir_size'] = -1
                    sourcedir['path_isdir'] = False

        existing_profile = None
        if overwrite_profile:
            existing_profile = BackupProfileModel.get_or_none(BackupProfileModel.name == self.name)
            if existing_profile:
                self._profile_dict['id'] = existing_profile.id
        if not overwrite_profile or not existing_profile:
            # Guarantee uniqueness of ids
            while BackupProfileModel.get_or_none(BackupProfileModel.id == self.id) is not None:
                self._profile_dict['id'] += 1

            # Add suffix in case names are the same
            if BackupProfileModel.get_or_none(BackupProfileModel.name == self.name) is not None:
                suffix = 1
                while BackupProfileModel.get_or_none(BackupProfileModel.name == f"{self.name}-{suffix}") is not None:
                    suffix += 1
                self._profile_dict['name'] = f"{self.name}-{suffix}"

        # Load existing repo or restore it
        if self._profile_dict['repo']:
            repo = RepoModel.get_or_none(RepoModel.url == self.repo_url)
            if repo is None:
                # Load repo from export
                repo = dict_to_model(RepoModel, self._profile_dict['repo'])
                repo.save(force_insert=True)
            self._profile_dict['repo'] = model_to_dict(repo)

        if self.repo_password:
            keyring.set_password('vorta-repo', self.repo_url, self.repo_password)
            del self._profile_dict['password']

        # Delete and recreate the tables to clear them
        if overwrite_settings:
            DB.drop_tables([SettingsModel, WifiSettingModel])
            DB.create_tables([SettingsModel, WifiSettingModel])
            SettingsModel.insert_many(self._profile_dict['SettingsModel']).execute()
            WifiSettingModel.insert_many(self._profile_dict['WifiSettingModel']).execute()

        # Set the profile ids to match new profile
        for source in self._profile_dict['SourceFileModel']:
            source['profile'] = self.id
        # Delete existing Sources to avoid duplicates
        SourceFileModel.delete().where(SourceFileModel.profile == self.id).execute()
        SourceFileModel.insert_many(self._profile_dict['SourceFileModel']).execute()

        # Restore ExclusionModel entries
        if 'ExclusionModel' in self._profile_dict:
            for exclusion in self._profile_dict['ExclusionModel']:
                exclusion['profile'] = self.id
            ExclusionModel.delete().where(ExclusionModel.profile == self.id).execute()
            ExclusionModel.insert_many(self._profile_dict['ExclusionModel']).execute()

        # Delete added dictionaries to make it match BackupProfileModel
        del self._profile_dict['SettingsModel']
        del self._profile_dict['SourceFileModel']
        del self._profile_dict['WifiSettingModel']
        del self._profile_dict['SchemaVersion']
        if 'ExclusionModel' in self._profile_dict:
            del self._profile_dict['ExclusionModel']

        # dict to profile
        new_profile = dict_to_model(BackupProfileModel, self._profile_dict)
        if overwrite_profile and existing_profile:
            force_insert = False
        else:
            force_insert = True
        new_profile.save(force_insert=force_insert)
        init_db()  # rerun db init code to perform the same operations on the new as as on application boot
        return new_profile

    @classmethod
    def from_json(cls, filename):
        with open(filename, 'r') as file:
            try:
                profile_export = ProfileExport(json.loads(file.read()))
            except JSONDecodeError as exception:
                raise ImportFailedException(
                    'This file does not contain valid JSON: {}'.format(str(exception))
                ) from exception
        return profile_export

    def to_json(self):
        return json.dumps(self._profile_dict, default=self._converter, indent=4)

    @staticmethod
    def _converter(obj):
        if isinstance(obj, datetime.datetime):
            return obj.__str__()


class VersionException(Exception):
    """For when current_version < export_version. Should only occur if downgrading"""

    pass


class ImportFailedException(Exception):
    """Raised when a profile could not be imported."""

    pass



import os
import shlex
import subprocess
import tempfile
from datetime import datetime as dt

from PyQt6.QtCore import QCoreApplication

from vorta import config
from vorta.i18n import trans_late, translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.store.models import ArchiveModel, RepoModel, SourceFileModel, WifiSettingModel
from vorta.utils import borg_compat, format_archive_name, get_network_status_monitor

from .borg_job import BorgJob


class BorgCreateJob(BorgJob):
    def process_result(self, result):
        if result['returncode'] in [0, 1] and 'archive' in result['data']:
            new_archive, created = ArchiveModel.get_or_create(
                snapshot_id=result['data']['archive']['id'],
                defaults={
                    'name': result['data']['archive']['name'],
                    # Convert to local time (for Borg 2.x UTC timestamps) before storing as naive datetime
                    'time': dt.fromisoformat(result['data']['archive']['start']).astimezone().replace(tzinfo=None),
                    'repo': result['params']['repo_id'],
                    'duration': result['data']['archive']['duration'],
                    'size': result['data']['archive']['stats'].get('deduplicated_size', 0),
                    'trigger': result['params'].get('category', 'user'),
                },
            )
            new_archive.save()
            if created and 'cache' in result['data'] and 'stats' in result['data']['cache']:
                stats = result['data']['cache']['stats']
                repo = RepoModel.get(id=result['params']['repo_id'])
                repo.total_size = stats['total_size']
                # repo.unique_csize = stats['unique_csize']
                repo.unique_size = stats['unique_size']
                repo.total_unique_chunks = stats['total_unique_chunks']
                repo.save()

            if result['returncode'] == 1:
                logs_link = link(config.LOG_DIR.as_uri(), translate('messages', 'logs'))
                message = format_richtext(
                    escape(translate('BorgCreateJob', 'Backup finished with warnings. See the %1 for details.')),
                    logs_link,
                )
                self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] " + message)
            else:
                self.app.backup_log_event.emit('', {})
                self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Backup finished.')}")

    def progress_event(self, fmt):
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {fmt}")

    def started_event(self):
        self.app.backup_started_event.emit()
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Backup started.')}")

    def finished_event(self, result):
        self.pre_post_backup_cmd(self.params, cmd='post_backup_cmd', returncode=result['returncode'])
        self.app.backup_finished_event.emit(result)
        self.result.emit(result)

    @classmethod
    def pre_post_backup_cmd(cls, params, cmd='pre_backup_cmd', returncode=0):
        cmd = getattr(params['profile'], cmd)
        if cmd:
            env = {
                **os.environ.copy(),
                'repo_url': params['repo'].url,
                'profile_name': params['profile'].name,
                'profile_slug': params['profile'].slug(),
                'returncode': str(returncode),
            }
            proc = subprocess.Popen(cmd, shell=True, env=env)

            while proc.poll() is None:
                # Preventing GUI freezing while running pre-backup command
                QCoreApplication.processEvents()

            return proc.returncode
        else:
            return 0  # 0 if no command was run.

    @classmethod
    def prepare(cls, profile):
        """
        `borg create` is called from different places and needs some preparation.
        Centralize it here and return the required arguments to the caller.
        """
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret
        else:
            ret['ok'] = False  # Set back to False, so we can do our own checks here.

        n_backup_folders = SourceFileModel.select().where(SourceFileModel.profile == profile).count()

        # cmd options like `--paths-from-command` require a command
        # that is appended to the arguments
        # $ borg create --paths-from-command repo::archive1 -- find /home/user -type f -size -76M
        extra_cmd_options = []
        suffix_command = []
        if profile.repo.create_backup_cmd:
            s1, sep, s2 = profile.repo.create_backup_cmd.partition('-- ')
            extra_cmd_options = shlex.split(s1)
            suffix_command = shlex.split(sep + s2)

        if n_backup_folders == 0 and '--paths-from-command' not in extra_cmd_options:
            ret['message'] = trans_late('messages', 'Add some folders to back up first.')
            return ret

        network_status_monitor = get_network_status_monitor()
        current_wifi = network_status_monitor.get_current_wifi()
        if current_wifi is not None:
            wifi_is_disallowed = WifiSettingModel.select().where(
                (WifiSettingModel.ssid == current_wifi)
                & (WifiSettingModel.allowed == False)  # noqa
                & (WifiSettingModel.profile == profile)
            )
            if wifi_is_disallowed.count() > 0 and profile.repo.is_remote_repo():
                ret['message'] = trans_late('messages', 'Current Wifi is not allowed.')
                ret['level'] = 'info'
                return ret

        if (
            profile.repo.is_remote_repo()
            and profile.dont_run_on_metered_networks
            and network_status_monitor.is_network_metered()
        ):
            ret['message'] = trans_late('messages', 'Not running backup over metered connection.')
            ret['level'] = 'info'
            return ret

        ret['profile'] = profile
        ret['repo'] = profile.repo

        # Run user-supplied pre-backup command
        if cls.pre_post_backup_cmd(ret) != 0:
            ret['message'] = trans_late('messages', 'Pre-backup command returned non-zero exit code.')
            return ret

        if not profile.repo.is_remote_repo() and not os.path.exists(profile.repo.url):
            ret['message'] = trans_late('messages', 'Repo folder not mounted or moved.')
            return ret

        if 'zstd' in profile.compression and not borg_compat.check('ZSTD'):
            ret['message'] = trans_late(
                'messages',
                'Your current Borg version does not support ZStd compression.',
            )
            return ret

        cmd = [
            'borg',
            'create',
            '--list',
            '--progress',
            '--info',
            '--log-json',
            '--json',
            '--filter=AM',
            '-C',
            profile.compression,
        ]
        cmd += extra_cmd_options

        # Function to extend command with exclude-if-present patterns
        if profile.exclude_if_present is not None:
            patterns = []
            for f in profile.exclude_if_present.split('\n'):
                f = f.strip()
                if f.startswith('[x]'):
                    patterns.append(f[3:].strip())  # Remove the '[x]' prefix

            for pattern in patterns:
                if pattern.strip():
                    cmd.extend(['--exclude-if-present', pattern])

        # Add excludes
        # Partly inspired by borgmatic/borgmatic/borg/create.py
        exclude_dirs = []
        for p in profile.get_combined_exclusion_string().split('\n'):
            if p.strip():
                expanded_directory = os.path.expanduser(p.strip())
                exclude_dirs.append(expanded_directory)

        if exclude_dirs:
            pattern_file = tempfile.NamedTemporaryFile('w', delete=True)
            pattern_file.write('\n'.join(exclude_dirs))
            pattern_file.flush()
            cmd.extend(['--exclude-from', pattern_file.name])
            ret['cleanup_files'].append(pattern_file)

        # Currently not in use, but may be added back to the UI later.
        # if profile.exclude_if_present is not None:
        #     for f in profile.exclude_if_present.split('\n'):
        #         if f.strip():
        #             cmd.extend(['--exclude-if-present', f.strip()])

        # Add repo url and source dirs.
        new_archive_name = format_archive_name(profile, profile.new_archive_name)

        if borg_compat.check('V2'):
            cmd += ["-r", profile.repo.url, new_archive_name]
        else:
            cmd.append(f"{profile.repo.url}::{new_archive_name}")

        for f in SourceFileModel.select().where(SourceFileModel.profile == profile.id):
            cmd.append(f.dir)

        cmd += suffix_command

        ret['message'] = trans_late('messages', 'Starting backup…')
        ret['ok'] = True
        ret['cmd'] = cmd

        return ret
